import numpy as np
import pytest

import cocos.device
import cocos.numerics as cn

# backend = 'cpu'
# backend = 'cuda'
# backend = 'opencl'
backend = None


test_data = [np.array([[1.0, -2.0],
                       [-3.0, 4.0]],
                      dtype=np.float32),
             np.array([[2, -3.0],
                       [9, 5.0]],
                      dtype=np.float32),
             np.array([[-100, 50],
                       [1, 30]],
                      dtype=np.float32),
             np.array([[20, 50],
                       [15, 22]],
                      dtype=np.float32)]


def compare_numpy_and_cocos(cocos_array: cn.ndarray, numpy_array: np.ndarray):
    return np.allclose(np.array(cocos_array).astype(bool), numpy_array)


@pytest.mark.parametrize("x", test_data)
def test_reciprocal(x):
    cocos.device.init(backend)
    cocos.device.info()

    # using numpy
    y = np.reciprocal(x)

    # using Cocos
    x_cocos = cn.array(x)
    y_cocos = cn.reciprocal(x_cocos)

    assert np.allclose(y_cocos, y)



import numpy as np
import sympy as sym

import cocos.numerics as cn
import cocos.symbolic as cs
import cocos.symbolic._lambdification


def test_lambdify():
    x = sym.Symbol('x')

    expr = sym.sin(x) + sym.cos(x)

    lambdified_cn = cocos.symbolic._lambdification.lambdify(x, expr)
    lambdified_sym = sym.lambdify(x, expr)

    x_vals = np.linspace(-np.pi, np.pi, 100, dtype=np.float32)
    x_vals_gpu = cn.array(x_vals)
    assert np.allclose(lambdified_sym(x_vals), lambdified_cn(x_vals_gpu))



import collections
from types import ModuleType
import typing as tp

import numpy as np
import sympy as sym

from cocos.symbolic import COCOS_TRANSLATIONS
from cocos import numerics as cn
from cocos.numerics.data_types import NumericArray
from cocos.utilities import check_and_make_sequence
from cocos.numerics.numerical_package_selector import \
    select_num_pack, \
    get_gpu_and_num_pack_by_dtype_from_iterable


################################################################################
# Works for both Matrices and Arrays
################################################################################
def _compute_replacement_functions(
        replacement_functions_cpu: tp.Tuple[tp.Callable, ...],
        replacement_functions_gpu: tp.Tuple[tp.Callable, ...],
        perform_cse: bool,
        state_vectors: tp.Tuple,
        t: float,
        gpu: bool,
        number_of_state_variables: int) \
        -> tp.Tuple[tp.Tuple, int]:

    R = max([state_argument.size
             for state_argument
             in state_vectors
             if isinstance(state_argument, (np.ndarray, cn.ndarray))] + [1])

    if number_of_state_variables != len(state_vectors):
        raise ValueError(f"The number of state vectors({len(state_vectors)}) "
                         f"does not match the number of variables in the "
                         f"function({number_of_state_variables}).")

    arguments = [t] + state_vectors
    if perform_cse:
        for k in range(len(replacement_functions_cpu)):
            if gpu:
                arguments.append(replacement_functions_gpu[k](*arguments))
            else:
                arguments.append(replacement_functions_cpu[k](*arguments))

    return arguments, R


################################################################################
# General Arrays
################################################################################
def lambdify_array_with_modules(
        symbols: tp.Tuple[sym.Symbol, ...],
        array_expression: tp.Union[sym.Array, sym.MatrixBase],
        numeric_time_functions: tp.Dict[str, tp.Callable],
        modules: tp.Tuple[str, ...]) \
        -> tp.Tuple[tp.Callable, ...]:

    if not isinstance(modules, list):
        modules = list(modules)

    if not (isinstance(array_expression, sym.Array) or
            isinstance(array_expression, sym.MatrixBase)):
        raise TypeError(f"symbolic_matrix_expression must be an instance of "
                        f"sympy.ArrayBase or sympy.MatrixBase "
                        f"but is of type {type(array_expression)}")

    n = np.asscalar(np.prod(array_expression.shape))
    result = []
    for i in range(n):
        index = tuple(np.unravel_index(i, array_expression.shape, order='F'))
        result.append(sym.lambdify(symbols,
                                   array_expression[index],
                                   modules=[numeric_time_functions] + modules))

    return tuple(result)


def lambdify_array(
        symbols: tp.Tuple[sym.Symbol, ...],
        array_expression: tp.Union[sym.Array, sym.MatrixBase],
        numeric_time_functions: tp.Optional[tp.Dict[str, tp.Callable]] = None) \
        -> tp.Tuple[tp.Tuple[tp.Callable, ...], tp.Tuple[tp.Callable, ...]]:

    if numeric_time_functions is None:
        numeric_time_functions = dict()

    functions_cpu = lambdify_array_with_modules(symbols,
                                                array_expression,
                                                numeric_time_functions,
                                                ('numpy', ))

    functions_gpu = lambdify_array_with_modules(symbols,
                                                array_expression,
                                                numeric_time_functions,
                                                (COCOS_TRANSLATIONS, ))

    return functions_cpu, functions_gpu


def _compute_result_internal(R: int,
                             dimensions: tp.Tuple[int, ...],
                             arguments,
                             functions_cpu: tp.Tuple[tp.Callable, ...],
                             functions_gpu: tp.Tuple[tp.Callable, ...],
                             pre_attach: bool,
                             gpu: bool,
                             dtype: np.generic) \
        -> NumericArray:

    num_pack = select_num_pack(gpu)
    if gpu:
        functions = functions_gpu
    else:
        functions = functions_cpu

    n = np.asscalar(np.prod(dimensions))

    if pre_attach:
        dimension_index = tuple([R] + list(dimensions))
    else:
        dimension_index = tuple(list(dimensions) + [R])

    result = num_pack.zeros(dimension_index, dtype=dtype)

    for i in range(n):
        index = tuple(np.unravel_index(i, dimensions, order='F'))

        function = functions[i]

        if pre_attach:
            location_index = tuple([slice(None)] + list(index))
        else:
            location_index = tuple(list(index) + [slice(None)])

        evaluated_function = function(*arguments)
        result[location_index] = evaluated_function

    return result


################################################################################
# Sequences of Lambdified Array, Matrix, and Vector-Expressions
################################################################################
class LambdifiedArrayExpressions(object):
    def __init__(
        self,
        argument_symbols: tp.Tuple[sym.Symbol],
        time_symbol: sym.Symbol,
        symbolic_array_expressions: tp.Tuple[tp.Union[sym.Array,
                                                      sym.Matrix],
                                             ...],
        numeric_time_functions: tp.Optional[tp.Dict[str, tp.Callable]] = None,
        squeeze_column_vectors: tp.Optional[tp.Tuple[bool, ...]] = None,
        perform_cse: bool = True,
        lazy_initialization: bool = False,
        pre_attach: tp.Optional[tp.Tuple[bool, ...]] = None,
        dtype: np.generic = np.float32):

        if numeric_time_functions is None:
            numeric_time_functions = dict()

        symbolic_array_expressions \
            = tuple(check_and_make_sequence(symbolic_array_expressions,
                                            sym.Matrix))

        if pre_attach is None:
            pre_attach = len(symbolic_array_expressions) * [True]
        pre_attach = check_and_make_sequence(pre_attach, bool)

        if squeeze_column_vectors is None:
            squeeze_column_vectors = len(symbolic_array_expressions) * [False]
        squeeze_column_vectors = check_and_make_sequence(squeeze_column_vectors,
                                                         bool)

        if not (len(symbolic_array_expressions) == len(pre_attach) and
                len(symbolic_array_expressions) == len(squeeze_column_vectors)):
            raise ValueError("symbolic_matrix_expressions, pre_attach, and "
                             "squeeze_column_vectors must have the same number "
                             "of elements")

        self._argument_symbols = argument_symbols
        self._time_symbol = time_symbol
        self._symbolic_array_expressions = symbolic_array_expressions
        self._numeric_time_functions = numeric_time_functions
        self._squeeze_column_vectors = squeeze_column_vectors
        self._symbols = tuple([time_symbol] + list(argument_symbols))

        self._perform_cse = perform_cse
        self._pre_attach = pre_attach
        self._dtype = dtype

        self._shapes \
            = tuple([symbolic_array_expression.shape
                     for symbolic_array_expression
                     in symbolic_array_expressions])

        if lazy_initialization:
            self._functions_cpu = None
            self._functions_gpu = None
        else:
            self._perform_initialization()

    def _perform_initialization(self):
        self._functions_cpu = []
        self._functions_gpu = []
        if self._perform_cse:
            repl, redu = sym.cse(self._symbolic_array_expressions,
                                 optimizations='basic')

            replacement_functions_cpu = []
            replacement_functions_gpu = []
            syms = [self._time_symbol] + list(self._argument_symbols)
            for i, v in enumerate(repl):
                replacement_functions_cpu.append(
                    sym.lambdify(syms,
                                 v[1],
                                 modules=[self._numeric_time_functions,
                                          'numpy']))

                replacement_functions_gpu.append(
                    sym.lambdify(
                        syms,
                        v[1],
                        modules=[self._numeric_time_functions,
                                 COCOS_TRANSLATIONS]))

                syms.append(v[0])

            self._replacement_functions_cpu = tuple(replacement_functions_cpu)
            self._replacement_functions_gpu = tuple(replacement_functions_gpu)

            for symbolic_matrix_expression in redu:
                functions_cpu, functions_gpu \
                    = lambdify_array(tuple(syms),
                                     symbolic_matrix_expression,
                                     self._numeric_time_functions)

                self._functions_cpu.append(functions_cpu)
                self._functions_gpu.append(functions_gpu)
        else:
            for symbolic_matrix_expression in self._symbolic_array_expressions:
                functions_cpu, functions_gpu \
                    = lambdify_array(self._symbols,
                                     symbolic_matrix_expression,
                                     self._numeric_time_functions)

                self._functions_cpu.append(functions_cpu)
                self._functions_gpu.append(functions_gpu)

    @property
    def shapes(self) -> tp.Tuple[tp.Tuple[int, ...], ...]:
        return self._shapes

    def is_column_vector(self, i: int) -> bool:
        shape = self._symbolic_array_expressions[i].shape
        return len(shape) == 1 \
               or all([dim == 1
                       for dim
                       in shape[1:]])

    def is_row_vector(self, i: int) -> bool:
        shape = self._symbolic_array_expressions[i].shape
        if len(shape < 2):
            return False
        else:
            return shape[0] == 1 and \
                   (len(shape) == 2 or
                    all([dim == 1
                         for dim
                         in shape[2:]]))

    @property
    def argument_symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._argument_symbols

    @property
    def number_of_state_variables(self) -> int:
        return len(self._argument_symbols)

    @property
    def time_symbol(self) -> sym.Symbol:
        return self._time_symbol

    @property
    def symbolic_array_expressions(self) -> tp.Tuple[sym.Matrix, ...]:
        return self._symbolic_array_expressions

    @property
    def numeric_time_functions(self) -> tp.Dict[str, tp.Callable]:
        return self._numeric_time_functions

    @property
    def symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._symbols

    @property
    def number_of_variables(self) -> int:
        return len(self._symbols)

    @property
    def dtype(self) -> np.generic:
        return self._dtype

    def evaluate_with_list_of_state_vectors(
            self,
            list_of_state_vectors: tp.Tuple[NumericArray, ...],
            t: float,
            gpu: bool = False) \
            -> tp.Tuple[NumericArray, ...]:

        if self._functions_cpu is None or self._functions_gpu is None:
            self._perform_initialization()

        arguments, R \
            = _compute_replacement_functions(self._replacement_functions_cpu,
                                             self._replacement_functions_gpu,
                                             self._perform_cse,
                                             list_of_state_vectors,
                                             t,
                                             gpu,
                                             self.number_of_state_variables)

        results = []
        for i, (shape,
                pre_attach,
                functions_cpu,
                functions_gpu,
                squeeze_column_vector) in \
                enumerate(zip(self.shapes,
                              self._pre_attach,
                              self._functions_cpu,
                              self._functions_gpu,
                              self._squeeze_column_vectors)):
            is_column_vector = self.is_column_vector(i)

            if is_column_vector and squeeze_column_vector:
                dimensions = (shape[0], )
            else:
                dimensions = shape

            result = _compute_result_internal(R,
                                              dimensions,
                                              arguments,
                                              functions_cpu,
                                              functions_gpu,
                                              pre_attach,
                                              gpu,
                                              self.dtype)
            results.append(result)

        return tuple(results)

    def evaluate(self,
                 state_matrices: tp.Tuple[NumericArray, ...],
                 t: float) \
            -> tp.Tuple[NumericArray, ...]:

        if not isinstance(state_matrices, collections.Sequence):
            if isinstance(state_matrices, (np.ndarray, cn.ndarray)):
                state_matrices = [state_matrices]
            else:
                raise TypeError("state_variables must be of type "
                                "Sequence[Union[numpy.ndarray, cocos.ndarray]]")

        gpu, num_pack \
            = get_gpu_and_num_pack_by_dtype_from_iterable(state_matrices)
        list_of_state_vectors = []
        for state_matrix in state_matrices:
            if state_matrix.ndim > 1:
                for i in range(state_matrix.shape[1]):
                    list_of_state_vectors.append((state_matrix[:, i]))
            else:
                list_of_state_vectors.append(state_matrix)

        return self.evaluate_with_list_of_state_vectors(list_of_state_vectors, t, gpu)


class LambdifiedMatrixExpressions(LambdifiedArrayExpressions):
    def __init__(
         self,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: sym.Symbol,
         symbolic_matrix_expressions: tp.Tuple[sym.Matrix, ...],
         numeric_time_functions: tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vectors: tp.Optional[tp.Tuple[bool, ...]] = None,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: tp.Optional[tp.Tuple[bool, ...]] = None,
         dtype: np.generic = np.float32):

        if numeric_time_functions is None:
            numeric_time_functions = dict()

        super().__init__(argument_symbols,
                         time_symbol,
                         symbolic_matrix_expressions,
                         numeric_time_functions,
                         squeeze_column_vectors,
                         perform_cse,
                         lazy_initialization,
                         pre_attach,
                         dtype)

        rows = []
        cols = []

        for symbolic_matrix_expression in symbolic_matrix_expressions:
            r, c = symbolic_matrix_expression.shape
            rows.append(r)
            cols.append(c)

        self._rows = tuple(rows)
        self._cols = tuple(cols)

    @property
    def rows(self) -> tp.Tuple[int, ...]:
        return self._rows

    @property
    def cols(self) -> tp.Tuple[int, ...]:
        return self._cols

    @property
    def symbolic_matrix_expressions(self) -> tp.Tuple[sym.Matrix, ...]:
        return self.symbolic_array_expressions


################################################################################
# Single Lambdified Array, Matrix, and Vector-Expressions
################################################################################
class LambdifiedArrayExpression(object):
    def __init__(
         self,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: sym.Symbol,
         symbolic_array_expression: tp.Union[sym.Matrix, sym.Array],
         numeric_time_functions: tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vector: bool = False,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        self._lambdified_array_expressions = \
            LambdifiedArrayExpressions(
                argument_symbols=argument_symbols,
                time_symbol=time_symbol,
                symbolic_array_expressions=(symbolic_array_expression, ),
                numeric_time_functions=numeric_time_functions,
                squeeze_column_vectors=(squeeze_column_vector, ),
                perform_cse=perform_cse,
                lazy_initialization=lazy_initialization,
                pre_attach=(pre_attach, ),
                dtype=dtype)

    @property
    def is_column_vector(self) -> bool:
        return self._lambdified_array_expressions.is_column_vector(0)

    @property
    def is_row_vector(self) -> bool:
        return self._lambdified_array_expressions.is_row_vector(0)

    @property
    def argument_symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._lambdified_array_expressions.argument_symbols

    @property
    def number_of_state_variables(self) -> int:
        return self._lambdified_array_expressions.number_of_state_variables

    @property
    def time_symbol(self) -> sym.Symbol:
        return self._lambdified_array_expressions.time_symbol

    @property
    def symbolic_array_expression(self) -> sym.Matrix:
        return self._lambdified_array_expressions.symbolic_array_expressions[0]

    @property
    def numeric_time_functions(self) -> tp.Dict[str, tp.Callable]:
        return self._lambdified_array_expressions.numeric_time_functions

    @property
    def symbols(self) -> tp.Tuple[sym.Symbol, ...]:
        return self._lambdified_array_expressions.symbols

    @property
    def number_of_variables(self) -> int:
        return self._lambdified_array_expressions.number_of_variables

    def evaluate_with_list_of_state_vectors(
            self,
            list_of_state_vectors: tp.Tuple[NumericArray, ...],
            t: float,
            gpu: bool = False) \
            -> NumericArray:

        return (self
                ._lambdified_array_expressions
                .evaluate_with_list_of_state_vectors(list_of_state_vectors,
                                                     t,
                                                     gpu)[0])

    def evaluate(self,
                 state_matrices: tp.Tuple[NumericArray, ...],
                 t: float) -> NumericArray:
        return self._lambdified_array_expressions.evaluate(state_matrices, t)[0]


class LambdifiedMatrixExpression(LambdifiedArrayExpression):
    def __init__(
         self,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: sym.Symbol,
         symbolic_matrix_expression: sym.Matrix,
         numeric_time_functions: tp.Optional[tp.Dict[str, tp.Callable]] = None,
         squeeze_column_vector: bool = False,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        super().__init__(argument_symbols,
                         time_symbol,
                         symbolic_matrix_expression,
                         numeric_time_functions,
                         squeeze_column_vector,
                         perform_cse,
                         lazy_initialization,
                         pre_attach,
                         dtype)

    @property
    def rows(self) -> int:
        return (self
                ._lambdified_array_expressions
                .symbolic_array_expressions[0]
                .shape[0])

    @property
    def cols(self) -> int:
        return (self
                ._lambdified_array_expressions
                .symbolic_array_expressions[0].shape[1])


class LambdifiedVectorExpression(LambdifiedMatrixExpression):
    def __init__(
         self,
         argument_symbols: tp.Tuple[sym.Symbol, ...],
         time_symbol: sym.Symbol,
         symbolic_vector_expression: sym.Matrix,
         numeric_time_functions: tp.Optional[tp.Dict[str, tp.Callable]] = None,
         perform_cse: bool = True,
         lazy_initialization: bool = False,
         pre_attach: bool = True,
         dtype: np.generic = np.float32):

        super().__init__(argument_symbols,
                         time_symbol,
                         symbolic_vector_expression,
                         numeric_time_functions,
                         True,
                         perform_cse,
                         lazy_initialization,
                         pre_attach,
                         dtype)


def lambdify(args,
             expr,
             modules: tp.Optional[tp.Tuple] = None,
             printer: tp.Optional[tp.Tuple] = None,
             use_imps: bool = True,
             dummify: bool = True):
    if modules is None:
        modules = []

    return sym.lambdify(args,
                        expr,
                        modules=list(modules) + [COCOS_TRANSLATIONS],
                        printer=printer,
                        use_imps=use_imps,
                        dummify=dummify)



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [(np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 20]],
                       dtype=np.int32), None),
             (np.array([[0.2, 1.0, 0.5],
                        [0.4, 0.5, 0.6],
                        [0.7, 0.2, 0.25]],
                       dtype=np.float32),
              np.array([0.2, 0.3, 0.5],
                       dtype=np.float32)),
             (np.array([[0.5, 2.3, 3.1],
                        [4, 5.5, 6],
                        [7 - 9j, 8 + 1j, 2 + 10j]],
                       dtype=np.complex64),
              np.array([0.2, 0.3, 0.5],
                       dtype=np.float32))]


@pytest.mark.parametrize("A_numpy, weights", test_data)
def test_average(A_numpy, weights):
    print(A_numpy)
    cocos.device.init()
    print('init')
    A_cocos = cn.array(A_numpy)

    if isinstance(weights, np.ndarray):
        weights_cocos = cn.array(weights)
    else:
        weights_cocos = None

    # conduct tests
    average_numpy = np.average(A_numpy)
    average_cocos = cn.average(A_cocos)
    assert np.allclose(average_numpy, average_cocos)

    average_numpy_axis_0 = np.average(A_numpy, axis=0, weights=weights)
    average_cocos_axis_0 = cn.average(A_cocos, axis=0, weights=weights_cocos)
    truth_value = np.allclose(average_numpy_axis_0, average_cocos_axis_0)
    if not truth_value:
        print("input array")
        print(A_numpy)
        print("input weights")
        print(weights)
        print("output numpy axis 0")
        print(average_numpy_axis_0)
        print("output cocos axis 0")
        print(average_cocos_axis_0)
    assert truth_value

    # ToDo: Taking the average over axis 1 with weights produces results that are identical to the operation without weights
    average_numpy_axis_1 = np.average(A_numpy, axis=1, weights=weights)
    average_cocos_axis_1 = cn.average(A_cocos, axis=1, weights=weights_cocos)
    truth_value = np.allclose(average_numpy_axis_1, average_cocos_axis_1)

    if not truth_value:
        print("input array")
        print(A_numpy)
        print("input weights")
        print(weights)
        print("output numpy axis 0")
        print(average_numpy_axis_0)
        print("output cocos axis 0")
        print(average_cocos_axis_0)

        print("output numpy axis 1")
        print(average_numpy_axis_1)
        print("output cocos axis 1")
        print(average_cocos_axis_1)

        print("output numpy axis 0 no weights")
        print(np.average(A_numpy, axis=0))

        print("output numpy axis 1 no weights")
        print(np.average(A_numpy, axis=1))
    assert truth_value



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [(2, 2), (3, 3)]


@pytest.mark.parametrize("n, m", test_data)
def test_identity(n: int, m: int):
    cocos.device.init()

    # eye
    A_numpy = np.eye(n, m)
    A_cocos = cn.eye(n, m)

    assert np.allclose(A_numpy, A_cocos)

    # identity
    B_numpy = np.identity(n)
    B_cocos = cn.identity(n)

    assert np.allclose(B_numpy, B_cocos)



import array as pyarray
import numpy as np
import pytest

import cocos.device
import cocos.numerics as cn


# backend = 'cpu'
# backend = 'cuda'
# backend = 'opencl'
backend = None


test_data = [(np.array([[1.0, -2.0], [-3.0, 4.0]], dtype=np.float32),
              np.array([[0.0, -3.0], [0.0, 5.0]], dtype=np.float32)),
             (np.array([[-100, 50], [1, 30]], dtype=np.float32),
              np.array([[20, 50], [15, 22]], dtype=np.float32))]


def compare_numpy_and_cocos(cocos_array: cn.ndarray, numpy_array: np.ndarray):
    return np.allclose(np.array(cocos_array).astype(bool), numpy_array)


@pytest.mark.parametrize("x, y", test_data)
def test_any_all(x, y):
    cocos.device.init(backend)
    cocos.device.info()

    # using numpy
    # x = np.array([[1.0, -2.0], [-3.0, 4.0]])
    # y = np.array([[0.0, -3.0], [0.0, 5.0]])
    z1 = x > y
    z2 = x < y
    z3 = x == y

    # print("numpy result")
    # print(z1)
    # print("numpy dtype = {}".format(z1.dtype))

    # using Cocos
    a = cn.array(x)
    b = cn.array(y)
    c1 = a > b
    c2 = a < b
    c3 = a == b

    # print("cocos result")
    # print(c1)
    # print("cocos dtype = {}".format(c1.dtype))

    # assert np.allclose(np.array(c1).astype(bool), z1)
    # assert np.allclose(np.array(cn.all(c1)).astype(bool), np.all(z1))
    # assert np.allclose(np.array(cn.all(c1, 0)).astype(bool), np.all(z1, 0))
    # assert np.allclose(np.array(cn.all(c1, 1)).astype(bool), np.all(z1, 1))

    list_of_comparisons = [(c1, z1), (c2, z2), (c3, z3)]

    for c, z in list_of_comparisons:
        assert compare_numpy_and_cocos(c, z)
        assert compare_numpy_and_cocos(cn.all(c), np.all(z))
        assert compare_numpy_and_cocos(cn.all(c, 0), np.all(z, 0))
        assert compare_numpy_and_cocos(cn.all(c, 1), np.all(z, 1))

        assert compare_numpy_and_cocos(cn.any(c), np.any(z))
        assert compare_numpy_and_cocos(cn.any(c, 0), np.any(z, 0))
        assert compare_numpy_and_cocos(cn.any(c, 1), np.any(z, 1))




from ._translations import COCOS_TRANSLATIONS

from ._lambdification import \
    lambdify, \
    LambdifiedArrayExpressions, \
    LambdifiedMatrixExpressions, \
    LambdifiedArrayExpression, \
    LambdifiedMatrixExpression, \
    LambdifiedVectorExpression



import array as pyarray
import arrayfire as af
import numbers
import numpy as np
import typing as tp

from cocos.options import \
    GPUOptions, \
    MixedComputationErrorLevel

from ._conversion import \
    convert_numpy_to_af_type, \
    convert_af_to_numpy_type

from ._utilities import \
    _as_str, \
    _pad_shape_tuple_none, \
    _pad_shape_tuple_axis, \
    _pad_shape_tuple_one, \
    _compute_slice_length, \
    _remove_trailing_ones


def _verify_operand(other):
    if  isinstance(other, np.ndarray):
        if GPUOptions.mixed_computation_error_level == \
                MixedComputationErrorLevel.ERROR:
            raise TypeError("Combining a gpu operation with a numpy array")
        elif GPUOptions.mixed_computation_error_level == \
                MixedComputationErrorLevel.WARNING:
            print("Warning: Combining a gpu operation with a numpy array "
                  "results in a conversion which is slow.")


def _binary_function(lhs, rhs, func) -> 'ndarray':
    _verify_operand(lhs)
    _verify_operand(rhs)

    if isinstance(lhs, ndarray):
        lhs = lhs._af_array

    if isinstance(rhs, ndarray):
        rhs = rhs._af_array

    new_af_array = func(lhs, rhs)
    return ndarray(new_af_array)


def _binary_method(rhs,
                   func,
                   broadcast: bool = False) -> 'ndarray':
    _verify_operand(rhs)
    argument = None
    if isinstance(rhs, ndarray):
        argument = rhs._af_array
    else:
        argument = rhs

    if broadcast:
        new_af_array = af.broadcast(func, argument)
    else:
        new_af_array = func(argument)
    return ndarray(new_af_array)


def _unary_function(x, func) -> 'ndarray':
    _verify_operand(x)
    # print("unary function argument is a {}".format(type(x)))
    if isinstance(x, ndarray):
        x = x._af_array
        return ndarray(func(x))
    else:
        return func(x)


def is_boolean(a):
    return a.dtype == np.bool or a.dtype == np.bool_


def _translate_index_key(item, input_shape: tp.Union[int, tp.Tuple[int, ...]]) \
        -> tp.Tuple[tp.Any, tp.Tuple[int, ...]]:
    if isinstance(item, tuple):
        # item is a tuple of items
        af_item = []
        required_shape = []

        for key, current_shape in zip(item, input_shape):
            translated_index_for_key, required_shape_for_key \
                = _translate_index_key(key, current_shape)
            af_item.append(translated_index_for_key)
            if isinstance(required_shape_for_key, tuple):
                required_shape_for_key = list(required_shape_for_key)
                required_shape += required_shape_for_key
            else:
                required_shape.append(required_shape_for_key)

        af_item = tuple(af_item)
        required_shape = tuple(required_shape)
    elif isinstance(item, ndarray):
        # item is a cocos array
        af_item = item._af_array
        if is_boolean(item):
            required_shape = (int(item.sum()), )
        else:
            required_shape = item.shape
    elif isinstance(item, np.ndarray):
        # item is a numpy array
        af_item = array(item)._af_array
        if is_boolean(item):
            required_shape = (int(item.sum()), )
        else:
            required_shape = item.shape
    elif isinstance(item, numbers.Number):
        af_item = item
        required_shape = (1, )
    elif isinstance(item, slice):
        af_item = item
        if isinstance(input_shape, tuple):
            input_shape = input_shape[0]
        required_shape = (_compute_slice_length(item, input_shape), )
    else:
        af_item = item
        required_shape = (-1,)

    return af_item, required_shape


# def _translate_index_key(item):
#     if isinstance(item, tuple):
#         af_item = []
#         for key in item:
#             if isinstance(key, ndarray):
#                 af_item.append(key._af_array)
#             else:
#                 af_item.append(key)
#
#         af_item = tuple(af_item)
#     elif isinstance(item, ndarray):
#         af_item = item._af_array
#     elif isinstance(item, np.ndarray):
#         af_item = array(item)
#     else:
#         af_item = item
#
#     return af_item


class ndarray(object):
    __array_priority__ = 35

    def __init__(self,
                 af_array: af.Array,
                 shape: tp.Optional[tp.Tuple[int]] = None):
        self._af_array = af_array
        self._label = 'array'
        self._shape = shape

    def __del__(self):
        del self._af_array

    def __getstate__(self) -> tp.Tuple:
        return np.array(self._af_array), self._label, self._shape

    def __setstate__(self, state: tp.Tuple):
        numpy_array, label, shape = state
        af_array = af.to_array(numpy_array)
        self._af_array = af_array
        self._label = label
        self._shape = shape

    def __array__(self) -> np.ndarray:

        return self._af_array.__array__()

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def dtype(self) -> np.generic:
        """
        Data-type of the array’s elements.
        """

        return convert_af_to_numpy_type(self._af_array.dtype())

    @property
    def shape(self) -> tp.Tuple[int]:
        """
        Tuple of array dimensions.
        """

        if self._shape is not None:
            return self._shape
        else:
            return self._af_array.dims()

    @property
    def ndim(self) -> int:
        """
        Number of array dimensions.
        """

        return len(self.shape)

    @property
    def size(self) -> int:
        """
        Number of elements in the array.
        """

        return self._af_array.elements()

    @property
    def nbytes(self) -> int:
        """
        Total bytes consumed by the elements of the array.
        """

        return self._af_array.allocated()

    @property
    def itemsize(self) -> int:
        """
        Length of one array element in bytes.
        """

        return int(self.nbytes / self.size)

    @property
    def strides(self):
        """
        Tuple of bytes to step in each dimension when traversing an array.
        """

        return self._af_array.strides()

    @property
    def T(self):
        """
        Same as self.transpose(), except that self is returned if self.ndim < 2.
        """

        if self.ndim < 2:
            return self
        else:
            return ndarray(af.transpose(self._af_array, False))

    @property
    def H(self):
        """
        Conjugate transpose.
        """

        return ndarray(af.transpose(self._af_array, True))

    @property
    def imag(self) -> 'ndarray':
        """
        The imaginary part of the array.
        """

        return imag(self)

    @property
    def real(self) -> 'ndarray':
        """
        The real part of the array.
        """

        return real(self)

    def astype(self,
               dtype: np.generic,
               order: str = 'K',
               casting: str = 'unsafe',
               subok: bool = True,
               copy: bool = True) \
            -> 'ndarray':
        """
        Copy of the array, cast to a specified type.
        """

        af_type = convert_numpy_to_af_type(dtype)
        af_array = self._af_array.as_type(af_type)
        return ndarray(af_array)

    def copy(self) -> 'ndarray':
        """
        Return a copy of the array.
        """

        af_array = self._af_array.copy()
        return ndarray(af_array)

    def __str__(self) -> str:
        if GPUOptions.str_via_numpy:
            np_array = np.array(self)
            return np_array.__str__()
        else:
            # return self._af_array._as_str()
            return _as_str(self._af_array, dims=True)

    def __repr__(self) -> str:
        np_array = np.array(self)
        return np_array.__repr__()

    def __len__(self) -> int:
        return self.shape[0]

    def __abs__(self):
        new_af_array = af.abs(self._af_array)
        return ndarray(new_af_array)

    @af.broadcast
    def __add__(self, other):
        return _binary_method(other, func=self._af_array.__add__)

    @af.broadcast
    def __iadd__(self, other):
        return _binary_method(other, func=self._af_array.__iadd__)

    @af.broadcast
    def __radd__(self, other):
        return _binary_method(other, func=self._af_array.__radd__)

    @af.broadcast
    def __sub__(self, other):
        return _binary_method(other, func=self._af_array.__sub__)

    @af.broadcast
    def __isub__(self, other):
        return _binary_method(other, func=self._af_array.__isub__)

    @af.broadcast
    def __rsub__(self, other):
        return _binary_method(other, func=self._af_array.__rsub__)

    @af.broadcast
    def __mul__(self, other):
        return _binary_method(other, func=self._af_array.__mul__)

    @af.broadcast
    def __imul__(self, other):
        return _binary_method(other, func=self._af_array.__imul__)

    @af.broadcast
    def __rmul__(self, other):
        return _binary_method(other, func=self._af_array.__rmul__)

    @af.broadcast
    def __truediv__(self, other):
        return _binary_method(other, func=self._af_array.__truediv__)

    @af.broadcast
    def __itruediv__(self, other):
        return _binary_method(other, func=self._af_array.__itruediv__)

    @af.broadcast
    def __rtruediv__(self, other):
        return _binary_method(other, func=self._af_array.__rtruediv__)

    @af.broadcast
    def __idiv__(self, other):
        return _binary_method(other, func=self._af_array.__idiv__)

    @af.broadcast
    def __rdiv__(self, other):
        return _binary_method(other, func=self._af_array.__rdiv__)

    @af.broadcast
    def __mod__(self, other):
        return _binary_method(other, func=self._af_array.__mod__)

    @af.broadcast
    def __imod__(self, other):
        return _binary_method(other, func=self._af_array.__imod__)

    @af.broadcast
    def __rmod__(self, other):
        return _binary_method(other, func=self._af_array.__rmod__)

    @af.broadcast
    def __pow__(self, other):
        return _binary_method(other, func=self._af_array.__pow__)

    @af.broadcast
    def __ipow__(self, other):
        return _binary_method(other, func=self._af_array.__ipow__)

    @af.broadcast
    def __rpow__(self, other):
        return _binary_method(other, func=self._af_array.__rpow__)

    @af.broadcast
    def __lt__(self, other):
        return _binary_method(other, func=self._af_array.__lt__)

    @af.broadcast
    def __gt__(self, other):
        return _binary_method(other, func=self._af_array.__gt__)

    @af.broadcast
    def __le__(self, other):
        return _binary_method(other, func=self._af_array.__le__)

    @af.broadcast
    def __ge__(self, other):
        return _binary_method(other, func=self._af_array.__ge__)

    @af.broadcast
    def __eq__(self, other):
        return _binary_method(other, func=self._af_array.__eq__)

    @af.broadcast
    def __ne__(self, other):
        return _binary_method(other, func=self._af_array.__ne__)

    @af.broadcast
    def __and__(self, other):
        return _binary_method(other, func=self._af_array.__and__)

    @af.broadcast
    def __iand__(self, other):
        return _binary_method(other, func=self._af_array.__iand__)

    @af.broadcast
    def __or__(self, other):
        return _binary_method(other, func=self._af_array.__or__)

    @af.broadcast
    def __ior__(self, other):
        return _binary_method(other, func=self._af_array.__ior__)

    @af.broadcast
    def __xor__(self, other):
        return _binary_method(other, func=self._af_array.__xor__)

    @af.broadcast
    def __ixor__(self, other):
        return _binary_method(other, func=self._af_array.__ixor__)

    @af.broadcast
    def __lshift__(self, other):
        return _binary_method(other, func=self._af_array.__lshift__)

    @af.broadcast
    def __ilshift__(self, other):
        return _binary_method(other, func=self._af_array.__ilshift__)

    @af.broadcast
    def __rshift__(self, other):
        return _binary_method(other, func=self._af_array.__rshift__)

    @af.broadcast
    def __irshift__(self, other):
        return _binary_method(other, func=self._af_array.__irshift__)

    def __neg__(self):
        return ndarray(self._af_array.__neg__())

    def __pos__(self):
        return ndarray(self._af_array.__pos__())

    def __invert__(self):
        return ndarray(self._af_array.__invert__())

    def __matmul__(self, other):
        _verify_operand(other)
        return dot(self, other)

    def __getitem__(self, item) -> 'ndarray':
        af_item, required_shape = _translate_index_key(item, self.shape)
        # from IPython.core.debugger import set_trace
        # set_trace()
        new_af_array = self._af_array.__getitem__(af_item)
        new_cocos_array = ndarray(new_af_array)

        if _remove_trailing_ones(new_cocos_array.shape) != \
                _remove_trailing_ones(required_shape):
            new_cocos_array = reshape_without_reorder(new_cocos_array,
                                                      newshape=required_shape)

        return new_cocos_array

    def __setitem__(self, key, value) -> 'ndarray':
        af_key, required_shape = _translate_index_key(key, self.shape)
        if isinstance(value, ndarray):
            value = value._af_array
        new_af_array = self._af_array.__setitem__(af_key, value)

        return ndarray(new_af_array)

    def __sizeof__(self):
        return self.size

    def dot(self,
            b: 'ndarray') -> 'ndarray':
        """
        Dot product of two arrays.
        """

        return dot(self, b)

    def _get_index_from_args(self, *args) \
            -> tp.Union[int, tp.Tuple[int, ...]]:
        if len(args) == 1 and isinstance(args[0], tuple):
            args = args[0]

        if len(args) == 1:
            if self.ndim == 1:
                index = args[0]
            else:
                index = np.unravel_index(args[0], self.shape)
        else:
            index = args

        return index

    def item(self, *args):
        """
        Copy an element of an array to a standard Python scalar and return it.
        """

        index = self._get_index_from_args(*args)
        af_item, required_shape = _translate_index_key(index, self.shape)
        return np.asscalar(np.array(self._af_array[af_item]))
        # return self._af_array[af_item].scalar()

    def itemset(self, *args):
        """
        Insert scalar into an array
        (scalar is cast to array’s dtype, if possible)
        """

        value = args[-1]
        index = self._get_index_from_args(*args[:-1])
        self[index] = value

    def conj(self) -> 'ndarray':
        """Complex-conjugate all elements."""

        return conj(self)

    def conjugate(self) -> 'ndarray':
        """
        Return the complex conjugate, element-wise.
        """

        return conj(self)

    def flatten(self) -> 'ndarray':
        """
        Return a copy of the array collapsed into one dimension.
        """

        af_array = af.data.flat(self._af_array)
        return ndarray(af_array)

    def reshape(self, shape: tp.Tuple[int]) -> 'ndarray':
        """
        Returns an array containing the same data with a new shape.
        """

        return reshape(self, shape)

    def squeeze(self,
                axis: tp.Optional[int] = None) -> 'ndarray':
        """
        Remove single-dimensional entries from the shape of a.
        """

        return squeeze(self, axis)

    def transpose(self):
        """
        Returns a new array with axes transposed.
        """

        return ndarray(af.transpose(self._af_array, False))

    def clip(self, min, max) -> 'ndarray':
        """
        Return an array whose values are limited to [min, max].
        """

        return ndarray(af.clamp(self._af_array, min, max))

    def max(self, axis: tp.Optional[int] = None):
        """
        Return the maximum along a given axis.
        """

        new_array = af.max(self._af_array, dim=axis)
        if isinstance(new_array, af.Array):
            return ndarray(new_array)
        else:
            return new_array

    def mean(self,
             axis: tp.Optional[int] = None,
             dtype: tp.Optional[np.generic] = None):
        """
        Returns the average of the array elements along given axis.
        """

        return mean(self, axis=axis)

    def min(self,
            axis: tp.Optional[int] = None):
        """
        Returns the minimum along given axis.
        """

        new_array = af.min(self._af_array, dim=axis)
        if isinstance(new_array, af.Array):
            return ndarray(new_array)
        else:
            return new_array

    def nonzero(self) -> tp.Tuple:
        """
        Return the indices of the elements that are non-zero.
        """

        return nonzero(self)

    def sort(self,
             axis: int = -1,
             ascending: bool = True):
        """
        Sort an array, in-place.
        """

        self._af_array = _sort_internal(self, axis, ascending)

    def all(self,
            axis: tp.Optional[int] = None) \
            -> tp.Union[bool, 'ndarray']:
        """
        Returns True if all elements evaluate to True.
        """

        return all(self, axis)

    def any(self,
            axis: tp.Optional[int] = None,
            out: tp.Optional = None,
            keepdims: bool = False) \
            -> tp.Union[bool, 'ndarray']:
        """
        Returns True if any of the elements of a evaluate to True.
        """

        return any(self, axis)

    def argmax(self,
               axis: tp.Optional[int] = None,
               out: tp.Optional['ndarray'] = None):
        """
        Return indices of the maximum values along the given axis.
        """

        val, idx = af.imax(self._af_array, dim=axis)
        return _wrap_af_array(idx)

    def argmin(self,
               axis: tp.Optional[int] = None,
               out: tp.Optional['ndarray'] = None):
        """
        Return indices of the minimum values along the given axis of a.
        """

        val, idx = af.imin(self._af_array, dim=axis)
        return _wrap_af_array(idx)

    def argsort(self,
                axis: int = -1,
                ascending: bool = True) \
            -> 'ndarray':
        """
        Returns the indices that would sort this array.
        """

        out, indices = sort_argsort(self, axis, ascending)
        return indices

    def cumsum(self,
               axis: tp.Optional[int] = None) \
            -> tp.Union[float, 'ndarray']:
        """
        Return the cumulative sum of the elements along the given axis.
        """

        return cumsum(self, axis)

    def diagonal(self,
                 offset: int = 0):
        """
        Return specified diagonals.
        """

        return ndarray(af.data.diag(self._af_array,
                                    num=offset,
                                    extract=True))

    def fill(self, value):
        """
        Fill the array with a scalar value.
        """

        shape = self.shape
        dtype = self.dtype
        del self._af_array
        self._af_array = _full_internal(shape, value, dtype)

    def prod(self,
             axis: tp.Optional[int] = None) \
            -> tp.Union[float, 'ndarray']:
        """
        Return the product of the array elements over the given axis
        """

        return prod(self)

    def ptp(self,
            axis: tp.Optional[int] = None) \
            -> 'ndarray':
        """
        Peak to peak (maximum - minimum) value along a given axis.
        """

        return ptp(self, axis)

    def repeat(self,
               repeats, axis: tp.Optional[int] = None) \
            -> 'ndarray':
        """
        Repeat elements of an array.
        """

        if axis is None:
            return tile(self.flatten(), repeats)
        else:
            if axis < 0 or axis >= self.ndim:
                raise ValueError(f"axis must be between 0 and {self.ndim-1}")
            new_shape = list(self.shape)
            new_shape[axis] *= repeats
            return tile(self.flatten(), repeats).reshape(tuple(new_shape))
            # reps = [1, 1, 1, 1]
            # reps[axis] = repeats
            # return tile(self, tuple(reps))

    def round(self) -> 'ndarray':
        return round(self)

    def std(self, axis: tp.Optional[int] = None) \
            -> tp.Union['ndarray', numbers.Number]:
        """
        Returns the standard deviation of the array elements along given axis.
        """

        return std(self, axis)

    def sum(self,
            axis: tp.Optional[int] = None) \
            -> tp.Union[numbers.Number, 'ndarray']:
        """
        Return the sum of the array elements over the given axis.
        """

        return sum(self, axis)

    def trace(self, offset: int = 0) -> numbers.Number:
        """
        Return the sum along diagonals of the array.
        """

        return trace(self, offset)

    def var(self,
            axis: tp.Optional[int] = None,
            ddof: int = 0) \
            -> tp.Union[numbers.Number, 'ndarray']:
        """
        Returns the variance of the array elements, along given axis.
        """

        return var(self, axis=axis, ddof=ddof)


def asscalar(a: ndarray) \
        -> numbers.Number:
    if a.size > 1:
        raise ValueError(f"a must be an array size 1 but has {a.size} elements")

    return np.asscalar(np.array(a))


def matmul(a: ndarray,
           b: ndarray) \
        -> ndarray:
    """
    Matrix product of two arrays.
    """

    return ndarray(af.matmul(a._af_array, b._af_array))


def copy(a: ndarray) -> ndarray:
    """
    Return an array copy of the given object.
    """

    return a.copy()


def ptp(a: ndarray,
        axis: tp.Optional[int] = None) -> ndarray:
    """
    Range of values (maximum - minimum) along an axis.
    """

    if axis is None:
        tmp = a.flatten()
    else:
        tmp = a

    return tmp.max(axis) - tmp.min(axis)


def array(object: tp.Union[tp.Sequence,
                           np.ndarray,
                           pyarray.array],
          dtype: tp.Optional[np.generic] = None,
          copy: bool = True,
          order: tp.Optional = None,
          subok: bool = False,
          ndmin: int = 0) \
        -> ndarray:
    """
    Create an array.
    """

    if order:
        raise ValueError('order != None is not supported')

    # if dtype is not None:
    #     raise ValueError("parameter dtype != None is not supported")

    af_array = None
    shape = None
    if isinstance(object, np.ndarray):
        if object.ndim == 0:
            # object = [np.asscalar(object)]
            object = np.reshape(object, (1,))
        af_array = af.to_array(object)
        shape = object.shape
    elif isinstance(object, list):
        np_array = np.array(object, dtype=dtype)
        af_array = af.to_array(np_array)
        shape = np_array.shape
    else:
        af_array = af.Array(object)

    result = ndarray(af_array, shape)
    if dtype is not None and result.dtype != dtype:
        result = result.astype(dtype)
    return result


def transpose(a: ndarray,
              axes: tp.Optional = None):
    """
    Permute the dimensions of an array.
    """

    if axes:
        raise ValueError('axes != None is not supported')

    return a.transpose()


def real(val: ndarray):
    """
    Return the real part of the complex argument.
    """

    return _unary_function(val, af.real)


def imag(a: ndarray):
    """
    Return the imaginary part of the complex argument.
    """

    return _unary_function(a, af.imag)


def isempty(num: ndarray) -> bool:
    return num._af_array.is_empty()


def isscalar(num: ndarray) -> bool:
    return num._af_array.is_scalar()


def isrow(num: ndarray) -> bool:
    return num._af_array.is_row()


def iscolumn(num: ndarray) -> bool:
    return num._af_array.is_column()


def isvector(num: ndarray) -> bool:
    return num._af_array.is_vector()


def isrealobj(x: ndarray) -> bool:
    return x._af_array.is_real()


def iscomplexobj(x: ndarray) -> bool:
    return x._af_array.is_complex()


def isdouble(num: ndarray) -> bool:
    return num._af_array.is_double()


def issingle(num: ndarray) -> bool:
    return num._af_array.is_single()


def isrealfloating(num: ndarray) -> bool:
    return num._af_array.is_real_floating()


def isfloating(num: ndarray) -> bool:
    return num._af_array.is_floating()


def isinteger(num: ndarray) -> bool:
    return num._af_array.is_integer()


def isbool(num: ndarray) -> bool:
    return num._af_array.is_integer()


def isfortran(a: ndarray) -> bool:
    if isinstance(a, ndarray):
        return True
    elif isinstance(a, np.ndarray):
        return np.isfortran(a)
    else:
        raise TypeError("argument must be an ndarray")


def display(array: ndarray):
    if isinstance(array, ndarray):
        af.display(array._af_array)
    else:
        print(array)


def _complete_shape(a: ndarray, shape: tp.Sequence[int]) -> tp.Tuple[
    int, ...]:
    shape_np = np.array(shape)
    placeholder_indices = np.where(shape_np == -1)[0]
    number_of_placeholders = len(placeholder_indices)

    assert number_of_placeholders >= 0
    if number_of_placeholders == 0:
        return tuple(shape)
    elif number_of_placeholders > 1:
        raise ValueError('can only specify one unknown dimension')
    elif number_of_placeholders == 1:
        position_of_placeholder = np.asscalar(placeholder_indices)
        dimensions_without_placeholder = shape_np[np.where(shape_np != -1)]
        size_of_all_other_dimensions = np.prod(dimensions_without_placeholder)
        remainder = a.size % size_of_all_other_dimensions
        if remainder != 0:
            raise ValueError(
                f'cannot reshape array of size {a.size} into shape {shape}')
        dimension_of_placeholder = a.size / size_of_all_other_dimensions
        shape_np[position_of_placeholder] = dimension_of_placeholder
        return tuple(shape_np)
    else:
        ValueError('This cannot happen')


def reorder(a: ndarray, new_order: tp.Tuple[int, ...]):
    if len(new_order) > 4:
        raise ValueError(f'Cocos does not support arrays with more than 4 axes.')

    d0, d1, d2, d3 = _pad_shape_tuple_axis(new_order)
    # print(f'd0={d0}, d1={d1}, d2={d2}, d3={d3}')

    output_array = af.reorder(a._af_array, d0, d1, d2, d3)
    return ndarray(output_array)


def reshape_without_reorder(a: ndarray,
                            newshape: tp.Tuple[int, ...]) -> ndarray:
    """
    Gives a new shape to an array without changing its data.
    """

    newshape = _complete_shape(a, newshape)
    input_array = a._af_array
    d0, d1, d2, d3 = _pad_shape_tuple_none(newshape)
    output_array = af.data.moddims(input_array, d0, d1, d2, d3)
    return ndarray(output_array)


def reshape(a, newshape):
    newshape = _complete_shape(a, newshape)
    if a.ndim < 1 or a.ndim > 4:
        raise ValueError('a must have between 1 and 4 axes')
    elif a.ndim == 1:
        tmp = a
    else:
        tmp = reorder(a, tuple(range(a.ndim))[::-1]).flatten()
    #     elif a.ndim == 2:
    #         tmp = cn.reorder(a_gpu, (1, 0)).flatten()
    #     elif a.ndim == 3:
    #         tmp = cn.reorder(a_gpu, (2, 1, 0)).flatten()

    tmp = reshape_without_reorder(tmp, newshape=newshape[::-1])
    return reorder(tmp, tuple(range(len(newshape)))[::-1])


def squeeze(a: ndarray,
            axis: tp.Optional[int] = None) -> ndarray:
    """
    Remove single-dimensional entries from the shape of an array.
    """

    shape = a.shape
    if axis is None:
        newshape = []
        for i in range(a.ndim):
            if shape[i] != 1:
                newshape.append(shape[i])
    else:
        if axis < 0 or axis >= a.ndim:
            raise ValueError(f"axis must be between 0 and {a.ndim-1}")

        if shape[axis] == 1:
            newshape = list(shape)
            del newshape[axis]
        else:
            raise ValueError(f"axis {axis} must have dimension 1 but has "
                             f"dimension {shape[axis]}")

    # print(f"newshape = {newshape}")
    newshape = tuple(newshape)
    if shape == newshape:
        return a
    else:
        return reshape(a, newshape)


def _dot_internal(a: ndarray, b: ndarray) -> ndarray:
    if (isrow(a) and isrow(b)) or (iscolumn(a) and iscolumn(b)):
        return _binary_function(a, b, af.dot)
    elif isrow(a) and iscolumn(b):
        # inner product
        return _binary_function(a, b, af.matmul)
    else:
        # outer product
        return _binary_function(a, b, af.matmul)


def dot(a: ndarray, b: ndarray) -> ndarray:
    """
    Dot product of two arrays.
    """

    if isvector(a) and isvector(b):
        return _dot_internal(a, b)
    else:
        result = _binary_function(a, b, af.matmul)
        if iscolumn(b):
            result = reshape(result, (result.size, 1))
        return result


def vdot(a: ndarray, b: ndarray) -> ndarray:
    """
    Return the dot product of two vectors.
    """

    if iscomplexobj(a):
        a = a.conj()

    if isvector(a) and isvector(b):
        return _dot_internal(a, b)
    elif a.size == b.size:
        if isvector(a):
            return _dot_internal(a, b.flatten())
        elif isvector(b):
            return _dot_internal(a.flatten(), b)
        else:
            return _dot_internal(a.flatten(), b.flatten())
    else:
        raise ValueError("a and b must have the same number of elements")


def multiply(x1, x2):
    """
    Multiply arguments element-wise.
    """

    _verify_operand(x1)
    _verify_operand(x2)

    return x1 * x2


def divide(x1, x2):
    """
    Divide arguments element-wise.
    """

    _verify_operand(x1)
    _verify_operand(x2)

    return x1 / x2


def average(a: ndarray,
            axis: tp.Optional[int] = None,
            weights: tp.Optional[ndarray] = None) \
        -> tp.Union[numbers.Number, ndarray]:
    """
    Compute the weighted average along the specified axis.
    """

    if weights is None:
        af_weights = None
    else:
        af_weights = weights._af_array

    new_af_array = af.mean(a._af_array,
                           weights=af_weights,
                           dim=axis)

    if isinstance(new_af_array, af.Array):
        return ndarray(new_af_array)
    else:
        return new_af_array


def mean(a: ndarray,
         axis: tp.Optional[int] = None) \
        -> tp.Union[numbers.Number, ndarray]:
    """
    Compute the arithmetic mean along the specified axis.
    """

    return average(a, axis=axis)


def median(a: ndarray,
           axis: tp.Optional[int] = None) \
        -> tp.Union[numbers.Number, ndarray]:
    """
    Compute the median along the specified axis.
    """

    new_af_array = af.median(a._af_array, dim=axis)
    if isinstance(new_af_array, af.Array):
        return ndarray(new_af_array)
    else:
        return new_af_array


def std(a: ndarray, axis: tp.Optional[int] = None, ddof: int = 0) \
        -> tp.Union[ndarray, numbers.Number]:
    """
    Compute the standard deviation along the specified axis.
    """

    new_af_array: tp.Optional[tp.Union[ndarray, numbers.Number]] = None
    if ddof or ddof == 0:
        new_af_array = af.stdev(a._af_array, dim=axis)
    elif ddof == 1:
        new_af_array = af.sqrt(af.var(a._af_array, isbiased=False, dim=axis))
    if isinstance(new_af_array, af.Array):
        return ndarray(new_af_array)
    else:
        return new_af_array


def var(a: ndarray,
        axis: tp.Optional[int] = None,
        dtype: tp.Optional[np.generic] = None,
        out: tp.Optional[ndarray] = None,
        ddof: int = 0,
        keepdims: bool = False) -> tp.Union[float, ndarray]:
    """
    Compute the variance along the specified axis.
    """

    isbiased = True
    if ddof is not None:
        if ddof == 1:
            isbiased = False
        elif ddof == 0:
            pass
        else:
            raise ValueError(f"ddof must be 0 or 1, ddof={ddof} is not "
                             f"supported")

    new_af_array \
        = af.var(a._af_array, isbiased=isbiased, weights=None, dim=axis)

    if isinstance(new_af_array, af.Array):
        return ndarray(new_af_array)
    else:
        return new_af_array


def cov(m: ndarray,
        y: tp.Optional[ndarray] = None,
        rowvar: bool = True,
        bias: bool = False,
        ddof: tp.Optional[int] = None,
        fweights: tp.Optional = None,
        aweights: tp.Optional = None):
    """
    Estimate a covariance matrix, given data and weights.
    """

    isbiased = bias
    if ddof is not None and ddof == 0:
        isbiased = True

    new_af_array = af.cov(m._af_array, isbiased=isbiased, dim=0)
    if isinstance(new_af_array, af.Array):
        return ndarray(new_af_array)
    else:
        return new_af_array


def corrcoef(x,
             y: tp.Optional = None,
             rowvar: int = 1,
             bias: bool = False,
             ddof: tp.Optional[int] = None):
    """
    Return Pearson product-moment correlation coefficients.
    """

    isbiased = bias
    if ddof and ddof == 0:
        isbiased = True

    cov_array = af.cov(x._af_array, isbiased=isbiased, dim=0)
    std_array = af.stdev(x._af_array, dim=0)

    if isinstance(std_array, af.Array) and isinstance(cov_array, af.Array):
        new_array = cov_array / af.matmulNT(std_array, std_array)
        return ndarray(new_array)
    elif (isinstance(std_array, af.Array) and
          not isinstance(cov_array, af.Array)) or \
            (not isinstance(std_array, af.Array) and
             isinstance(cov_array, af.Array)):
        raise ValueError("error: either both std_array and cov_array must be "
                         "arrayfire arrays or none of them")
    else:
        return cov_array / std_array * std_array


def _division_with_remainder(dividend: ndarray,
                             divisor_int: int) \
        -> tp.Tuple[ndarray, ndarray]:

    # divisor_array = array(divisor_int)
    whole_part = dividend / divisor_int
    remainder = dividend % divisor_int
    return whole_part, remainder


def nonzero(a: ndarray) -> tp.Tuple:
    index_af_array = af.where(a._af_array)

    tmp_array = ndarray(index_af_array)

    if a.ndim == 1:
        return (tmp_array, )
        # return tmp_array
    elif a.ndim == 2:
        return _division_with_remainder(tmp_array, a.shape[0])
    elif a.ndim == 3:
        tmp_array2, dimension2 = _division_with_remainder(tmp_array, a.shape[1])
        dimension0, dimension1 \
            = _division_with_remainder(tmp_array2, a.shape[0])

        return dimension0, dimension1, dimension2
    else:
        raise ValueError("function nonzero is not implemented for arrays with "
                         "more than three axes")


def trace(a: ndarray,
          offset: int=0) \
        -> tp.Union[int, float]:

    return af.sum(af.diag(a._af_array, num=offset))


def reciprocal(x: ndarray) -> ndarray:
    """
    Return the reciprocal of the argument, element-wise.
    """

    return 1.0 / x


def negative(x: ndarray) -> ndarray:
    """
    Numerical negative, element-wise.
    """

    return -x


def _wrap_af_array(af_array) -> ndarray:
    if isinstance(af_array, af.Array):
        return ndarray(af_array)
    else:
        return af_array


def _sort_internal(a: ndarray,
                   axis: int = -1,
                   ascending: bool = True) \
        -> af.Array:
    if axis is None:
        a = a.flatten()
        axis = 0
    elif axis == -1:
        axis = a.ndim - 1
    elif axis >= a.ndim:
        raise ValueError(f"Parameter axis must be between -1 and {a.ndim - 1}")

    return af.sort(a._af_array, dim=axis, is_ascending=ascending)


def all(a: ndarray,
        axis: tp.Optional[int] = None,
        out: tp.Optional[ndarray] = None,
        keepdims: bool = False) \
        -> tp.Union[bool, ndarray]:
    """
    Returns True if all elements evaluate to True.
    """

    if out:
        raise ValueError('out != None is not supported')

    return _wrap_af_array(af.all_true(a._af_array, dim=axis))


def any(a: ndarray,
        axis: tp.Optional[int] = None,
        out: tp.Optional[ndarray] = None,
        keepdims: bool = False) \
        -> tp.Union[bool, ndarray]:
    return _wrap_af_array(af.any_true(a._af_array, dim=axis))


def argmax(a: ndarray,
           axis: tp.Optional[int] = None,
           out: tp.Optional[ndarray] = None) -> ndarray:
    if out:
        raise ValueError('out != None is not supported')

    val, idx = af.imax(a._af_array, dim=axis)
    return _wrap_af_array(idx)


def argmin(a: ndarray,
           axis: tp.Optional[int] = None,
           out: tp.Optional[ndarray] = None) -> ndarray:
    if out:
        raise ValueError('out != None is not supported')

    val, idx = af.imin(a._af_array, dim=axis)
    return _wrap_af_array(idx)


def argsort(a: ndarray,
            axis: int = -1,
            ascending: bool = True) -> ndarray:
    out, indices = sort_argsort(a, axis, ascending)
    return indices


def conj(x: ndarray):
    """
    Return the complex conjugate, element-wise.
    """

    return _unary_function(x, af.conjg)


def cumsum(a: ndarray,
           axis: tp.Optional[int] = None,
           dtype: tp.Optional[np.generic] = None,
           out: tp.Optional[ndarray] = None) \
        -> tp.Union[float, ndarray]:
    """
    Return the cumulative sum of the elements along a given axis.
    """

    if axis is None:
        flat_array = a.flatten()
        new_af_array = af.accum(flat_array._af_array)
    else:
        new_af_array = af.accum(a._af_array, dim=axis)

    return ndarray(new_af_array)


def prod(a: ndarray, axis: tp.Optional[int] = None) -> tp.Union[float, ndarray]:
    """
    Return the product of array elements over a given axis.
    """

    return _wrap_af_array(af.product(a._af_array, dim=axis))


def sort(a: ndarray, axis: int = -1, ascending: bool = True) -> ndarray:
    return ndarray(_sort_internal(a, axis, ascending))


def sort_argsort(a: ndarray, axis: int = -1, ascending: bool = True) \
        -> tp.Tuple[ndarray, ndarray]:
    if axis is None:
        a = a.flatten()
        axis = 0
    elif axis == -1:
        axis = a.ndim - 1
    elif axis >= a.ndim:
        raise ValueError(f"Parameter axis must be between -1 and {a.ndim - 1}")

    af_out_array, af_idx_array \
        = af.sort_index(a._af_array, dim=axis, is_ascending=ascending)

    return ndarray(af_out_array), ndarray(af_idx_array)


def sum(a: ndarray, axis: tp.Optional[int] = None) \
        -> tp.Union[numbers.Number, ndarray]:
    """
    Sum of array elements over a given axis.
    """

    return _wrap_af_array(af.sum(a._af_array, dim=axis))


# from ._data
def _full_internal(shape: tp.Tuple[int],
                   value,
                   dtype: np.generic = np.float32) -> af.Array:
    af_type = convert_numpy_to_af_type(dtype)

    d0, d1, d2, d3 = _pad_shape_tuple_none(shape)
    return af.data.constant(value, d0=d0, d1=d1, d2=d2, d3=d3, dtype=af_type)


def full(shape: tp.Tuple[int], fill_value, dtype: np.generic = np.float32) \
        -> ndarray:
    """
    Return a new array of given shape and type, filled with fill_value.
    """

    return ndarray(_full_internal(shape, fill_value, dtype))


def full_like(a, fill_value, dtype: tp.Optional[np.generic] = None) -> ndarray:
    """
    Return a full array with the same shape and type as a given array.
    """

    if dtype is None:
        dtype = a.dtype
    return full(a.shape, fill_value, dtype)


def tile(A: ndarray, reps: tp.Union[int, tp.Tuple[int, ...]]) -> ndarray:
    af_array = None
    if isinstance(reps, tuple):
        d0, d1, d2, d3 = _pad_shape_tuple_one(reps)
        af_array = af.data.tile(A._af_array, d0, d1, d2, d3)
    elif isinstance(reps, int):
        if not reps > 0:
            raise ValueError("reps must be strictly positive")
        af_array = af.data.tile(A._af_array, reps)
    else:
        raise TypeError("types supported for reps are int an Tuple")

    return ndarray(af_array)


# from ._arith
def round(a: ndarray):
    """
    Return a with each element rounded to the given number of decimals.
    """

    return _unary_function(a, af.round)


def array_equal(a1: ndarray, a2: ndarray) -> bool:
    return all(a1 == a2)


def greater(x1: ndarray, x2: ndarray) -> ndarray:
    return x1 > x2


def greater_equal(x1: ndarray, x2: ndarray) -> ndarray:
    return x1 >= x2


def less(x1: ndarray, x2: ndarray) -> ndarray:
    return x1 < x2


def less_equal(x1: ndarray, x2: ndarray) -> ndarray:
    return x1 <= x2


def equal(x1: ndarray, x2: ndarray) -> ndarray:
    return x1 == x2


def not_equal(x1: ndarray, x2: ndarray) -> ndarray:
    return x1 != x2


def add(x1: ndarray, x2: ndarray) -> ndarray:
    """
    Add arguments element-wise.
    """

    return x1 + x2


def subtract(x1: ndarray, x2: ndarray) -> ndarray:
    """
    Subtract arguments, element-wise.
    """

    return x1 - x2


def true_divide(x1: ndarray, x2: ndarray) -> ndarray:
    """
    Returns a true division of the inputs, element-wise.
    """

    return x1 / x2


def floor(a: ndarray):
    """
    Return the floor of the input, element-wise.
    """

    return _unary_function(a, af.floor)


def floor_divide(x1: ndarray, x2: ndarray) -> ndarray:
    """
    Return the largest integer smaller or equal to the division of the inputs.
    """

    return floor(x1 / x2)


def clip(a: ndarray, a_min, a_max) -> 'ndarray':
    """
    Clip (limit) the values in an array.
    """

    return a.clip(a_min, a_max)



import numpy as np
import pytest
import typing as tp

from cocos.numerics._array import _complete_shape

two_by_two = np.array([[1, 2], [3, 4]])
two_by_two_by_two = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

test_data = [(two_by_two, (1, -1), (1, 4)),
             (two_by_two, (2, -1), (2, 2)),
             (two_by_two, (4, -1), (4, 1)),
             (two_by_two, (-1, 1), (4, 1)),
             (two_by_two, (-1, 2), (2, 2)),
             (two_by_two, (-1, 4), (1, 4)),
             (two_by_two_by_two, (-1, 2, 2), (2, 2, 2)),
             (two_by_two_by_two, (-1, 1, 1), (8, 1, 1)),
             (two_by_two_by_two, (-1, 2), (4, 2)), ]


@pytest.mark.parametrize("a, input_shape, intended_output_shape", test_data)
def test_complete_shape(a,
                        input_shape: tp.Tuple[int, ...],
                        intended_output_shape: tp.Tuple[int, ...]):

    output_shape = _complete_shape(a, input_shape)
    assert output_shape == intended_output_shape
    # y = x[index]
    # computed_length = _compute_slice_length(index, len(x))
    # if not len(y) == computed_length:
    #     print(f'computed length = {computed_length}, actual length = {len(y)}')
    # assert len(y) == computed_length



import arrayfire as af
from arrayfire.library import Dtype
import collections
import math
import numpy as np
from types import ModuleType
import typing as tp
from ._arith import \
    exp, \
    log, \
    minimum, \
    sqrt, \
    minimum

from ._array import asscalar, ndarray
from ._conversion import \
    convert_numpy_to_af_type, \
    convert_af_to_numpy_type

from cocos.numerics.linalg import cholesky

from cocos.options import \
    GPUOptions, \
    RandomNumberGenerator


SIZE_TYPE = tp.Optional[tp.Union[int, tp.Sequence]]


def map_rng_to_random_engine(rng: RandomNumberGenerator):
    if rng == RandomNumberGenerator.PHILOX_4X32_10:
        return af.random.RANDOM_ENGINE.PHILOX_4X32_10
    elif rng == RandomNumberGenerator.THREEFRY_2X32_16:
        return af.random.RANDOM_ENGINE.THREEFRY_2X32_16
    elif rng == RandomNumberGenerator.MERSENNE_GP11213:
        return af.random.RANDOM_ENGINE.MERSENNE_GP11213
    elif rng == RandomNumberGenerator.PHILOX:
        return af.random.RANDOM_ENGINE.PHILOX
    elif rng == RandomNumberGenerator.THREEFRY:
        return af.random.RANDOM_ENGINE.THREEFRY
    elif rng == RandomNumberGenerator.DEFAULT:
        return af.random.RANDOM_ENGINE.DEFAULT
    else:
        raise ValueError("The requested random number generator "
                         "is not supported.")


# initialized default random number engine
random_engine \
    = af.random.Random_Engine(
        engine_type=map_rng_to_random_engine(GPUOptions.default_rng))


################################################################################
# functions to get and set the seed
################################################################################

def seed(seed: tp.Optional[int] = None):
    """
    Seed the generator.
    """

    if seed is None:
        seed = 0
    af.set_seed(seed)


def get_seed() -> int:
    """
    Returns the current seed of the generator.
    """

    return af.get_seed()


################################################################################
# supporting functions for antithetic random numbers
################################################################################

def get_antithetic_slices(shape: tp.Sequence[int],
                          antithetic_dimension: int) \
        -> tp.Tuple[slice, ...]:
    """
    This function generates a tuple of slices to index the original array of
    random numbers to take either half or one less than half of the the original
    random numbers along the antithetic dimension and all of the random numbers
    along the other dimensions.
    """

    slices = []

    for axis, dimension in enumerate(shape):
        if axis == antithetic_dimension:
            s = slice(0, math.floor(dimension / 2), 1)
        else:
            s = slice(0, dimension, 1)

        slices.append(s)

    return tuple(slices)


def verify_shape_and_antithetic_dimension(shape: tp.Sequence[int],
                                          antithetic_dimension: tp.Optional[
                                              int] = None):
    """
    This function makes sure that the length shape argument is between 1 and 4
    and checks that the antithetic dimension is one of the dimensions in the
    shape argument.
    """

    if len(shape) > 4:
        raise ValueError('arrays with more than 4 axes are not supported')

    if len(shape) < 1:
        raise ValueError('array must have at least one axis')

    if antithetic_dimension < 0 or antithetic_dimension > len(shape) - 1:
        raise ValueError(
            f'antithetic dimension must be None or between 0 and {len(shape)}')


################################################################################
# Basic continuous random number generators
################################################################################

def rand(d0: int,
         d1: tp.Optional[int] = None,
         d2: tp.Optional[int] = None,
         d3: tp.Optional[int] = None,
         dtype: np.generic = np.float32) -> ndarray:
    """
    Random values in a given shape.
    """

    af_type = convert_numpy_to_af_type(dtype)
    af_array = af.data.randu(d0, d1, d2, d3, dtype=af_type)

    return ndarray(af_array)


def randn(d0: int,
          d1: tp.Optional[int] = None,
          d2: tp.Optional[int] = None,
          d3: tp.Optional[int] = None,
          dtype: np.generic = np.float32):
    """
    Return a sample (or samples) from the “standard normal” distribution.
    """

    af_type = convert_numpy_to_af_type(dtype)
    af_array = af.data.randn(d0, d1, d2, d3, dtype=af_type)
    return ndarray(af_array)


def randn_antithetic(shape: tp.Sequence[int],
                     antithetic_dimension: tp.Optional[int] = None,
                     dtype: np.generic = np.float32,
                     num_pack: ModuleType = np):
    verify_shape_and_antithetic_dimension(shape, antithetic_dimension)
    draw_shape = list(shape)

    if antithetic_dimension is not None:
        # adjust dimension over which antithetic random numbers are to be drawn
        draw_shape[antithetic_dimension] \
            = math.ceil(shape[antithetic_dimension] / 2)

    # draw original random numbers
    if num_pack == np:
        z = num_pack.random.randn(*draw_shape)
    else:
        z = num_pack.random.randn(*draw_shape, dtype=dtype)

    if antithetic_dimension is not None:
        # reflect random numbers at 0 and concatenate to original random numbers
        slices = get_antithetic_slices(shape, antithetic_dimension)
        z = num_pack.concatenate((z, -z[slices]),
                                 axis=antithetic_dimension)

    return z


def rand_antithetic(shape: tp.Sequence[int],
                    antithetic_dimension: tp.Optional[int] = None,
                    dtype: np.generic = np.float32,
                    num_pack: ModuleType = np):
    verify_shape_and_antithetic_dimension(shape, antithetic_dimension)
    draw_shape = list(shape)

    if antithetic_dimension is not None:
        # adjust dimension over which antithetic random numbers are to be drawn
        draw_shape[antithetic_dimension] \
            = math.ceil(shape[antithetic_dimension] / 2)

    # draw original random numbers
    if num_pack == np:
        u = num_pack.random.rand(*draw_shape)
    else:
        u = num_pack.random.rand(*draw_shape, dtype=dtype)

    if antithetic_dimension is not None:
        # reflect random numbers at 0 and concatenate to original random numbers
        slices = get_antithetic_slices(shape, antithetic_dimension)
        u = num_pack.concatenate((u, 1.0 - u[slices]),
                                 axis=antithetic_dimension)

    return u


################################################################################
# Basic discrete random number generators
################################################################################

def randint(low: int,
            high: tp.Optional[int] = None,
            size: tp.Optional[tp.Union[tp.Tuple[int, ...], int]] = None,
            dtype: np.generic = np.int32) \
        -> ndarray:
    """
    Draws an array of random integers ranging from low to high-1 of the
    specified shape.

    :param low: lowest number to draw
    :param high: highest integer to draw (excluding)
    :param size: shape of output array
    :param dtype: data type of integer to be generated
    :return: an ndarray of random integers
    """
    if not high:
        high = low
        low = 0

    if not size:
        size = (1,)
    elif isinstance(size, int):
        size = (size,)

    n = np.prod(size)
    divisor = 1.0 / (high - low)

    u = rand(n)
    u = minimum(u, 1.0 - np.finfo(np.float32).eps)
    if dtype != np.int32:
        i = (u / divisor).astype(np.int64) + low
        i = i.astype(dtype)
    else:
        i = (u / divisor).astype(np.int32) + low

    return i.reshape(size)


def choice(a: ndarray,
           size: tp.Optional[tp.Union[tp.Tuple[int, ...], int]] = None,
           replace: bool = True,
           p: tp.Optional[ndarray] = None) -> ndarray:
    if p:
        raise ValueError('p != None is not supported')

    if not replace:
        raise ValueError('replace=False is not supported')

    i = randint(0, a.size, size=size)

    if not isinstance(size, int):
        return a[i].reshape(size)
    else:
        return a[i]


def _draw_and_reshape(size: SIZE_TYPE,
                      rng_func: tp.Callable[[int], ndarray]) \
        -> ndarray:
    if not size:
        n = 1
    elif isinstance(size, int):
        n = size
    elif isinstance(size, (list, tuple)):
        n = np.prod(size)
    else:
        raise TypeError("size must be either of type int or tuple")

    random_numbers = rng_func(n)

    if size is None:
        random_numbers = asscalar(random_numbers)
    elif not isinstance(size, int):
        random_numbers = random_numbers.reshape(size)

    return random_numbers


def uniform(low: float = 0.0,
            high: float = 1.0,
            size: tp.Optional[SIZE_TYPE] = None):
    """
    Draw samples from a uniform distribution.
    """

    if high < low:
        raise ValueError("high must not be less than low")

    u = _draw_and_reshape(size, rand)
    return u * (high - low) + low


def _exponential_internal(scale: float,
                          n: int,
                          antithetic: bool = False) -> ndarray:
    u = rand(n)
    u = minimum(u, 1.0 - np.finfo(np.float32).eps)
    x: ndarray = log(1.0 - u) * (-scale)
    return x


def exponential(scale: float=1.0,
                size: tp.Optional[SIZE_TYPE] = None,
                antithethic: bool = False) -> ndarray:
    return _draw_and_reshape(size,
                             lambda n: _exponential_internal(
                                            scale=scale,
                                            n=n,
                                            antithetic=antithethic))


def standard_exponential(size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return exponential(size=size)


################################################################################
# gamma random number generator by Marsaglia and Tsang
# using Cocos vectorization
################################################################################

def gamma_rand_marsaglia_and_tsang_arrayfire(alpha: float,
                                             lambda_: float,
                                             n: int) \
        -> af.array:
    random_numbers = af.constant(0, n, dtype=Dtype.f32)
    # Gamma(alpha, lambda) generator using Marsaglia and Tsang method
    # Algorithm 4.33
    if alpha >= 1.0:
        d = alpha - 1 / 3
        c = 1.0 / np.sqrt(9.0 * d)

        number_generated = 0
        number_generated_total = 0

        while number_generated < n:
            number_left = n - number_generated

            z = af.randn(number_left, dtype=Dtype.f32)
            y = (1.0 + c * z)
            v = y * y * y

            accept_index_1 = ((z >= -1.0 / c) & (v > 0.0))
            z_accept_1 = z[accept_index_1]
            # del z
            v_accept_1 = v[accept_index_1]
            # del v
            u_accept_1 = af.randu(v_accept_1.elements(), dtype=Dtype.f32)
            # del U

            accept_index_2 = \
                u_accept_1 < af.exp((0.5 * z_accept_1 * z_accept_1 + d - d * v_accept_1 + d * af.log(v_accept_1)))

            x_accept = d * v_accept_1[accept_index_2] / lambda_
            number_accept = x_accept.elements()

            random_numbers[number_generated:np.minimum(n, number_generated + number_accept)] = \
                x_accept[0:np.minimum(number_left, number_accept)]

            number_generated += number_accept
            number_generated_total += number_left

        if GPUOptions.verbose:
            print(f"Acceptance ratio = {n/number_generated_total}")
    else:
        random_numbers = gamma_rand_marsaglia_and_tsang_arrayfire(alpha + 1, lambda_, n)
        random_numbers *= af.randu(n, dtype=Dtype.f32) ** (1.0 / alpha)

    return random_numbers


def gamma(shape: float,
          scale: float = 1.0,
          size: tp.Optional[SIZE_TYPE] = None) \
        -> ndarray:
    def fun(n: int):
        return ndarray(gamma_rand_marsaglia_and_tsang_arrayfire(
                            alpha=shape,
                            lambda_=1.0/scale, n=n))

    return _draw_and_reshape(size, lambda n: fun(n))


def standard_gamma(shape: float,
                   size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return gamma(shape, size=size)


def chisquare(df, size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return gamma(df / 2.0, 2.0, size)


def _beta_internal(a: float,
                   b: float,
                   n: int) -> ndarray:
    X = gamma(a, 1.0, n)
    Y = gamma(b, 1.0, n)
    return X / (X + Y)


def beta(a: float,
         b: float,
         size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return _draw_and_reshape(size, lambda n: _beta_internal(a, b, n))


def _wald_internal(mu: float,
                   LAMBDA: float,
                   n: int) -> ndarray:
    v = randn(n)
    u = rand(n)

    y = v * v
    del v
    x = mu + mu ** 2 / (2.0 * LAMBDA) * y - mu / (2.0 * LAMBDA) * sqrt(4.0 * mu * LAMBDA * y + mu ** 2.0 * y * y)
    reject_index = u > (mu / (mu + x))
    x[reject_index] = mu ** 2 / x[reject_index]
    return x


def wald(mean: float,
         scale: float,
         size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return _draw_and_reshape(size, lambda n: _wald_internal(mean, scale, n))


def normal(loc: float = 0.0,
           scale: float = 1.0,
           size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return _draw_and_reshape(size, lambda n: loc + scale * randn(n))


def standard_normal(size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return _draw_and_reshape(size, randn)


def lognormal(mean: float = 0.0,
              sigma: float = 1.0,
              size: tp.Optional[SIZE_TYPE] = None) -> ndarray:
    return exp(normal(mean, sigma, size))


def _logistic_internal(loc: float,
                       scale: float,
                       n: int) -> ndarray:
    u = rand(n)
    u = minimum(u, 1.0 - np.finfo(np.float32).eps)
    x: ndarray = loc - scale * log(1.0 / u - 1.0)
    return x


def logistic(loc: float = 0.0,
             scale: float = 1.0,
             size: tp.Optional[SIZE_TYPE] = None):
    return _draw_and_reshape(size, lambda n: _logistic_internal(loc, scale, n))


def multivariate_normal(mean, cov, size: tp.Sequence[int]) \
        -> ndarray:
    d = len(mean)
    if not isinstance(size, collections.Iterable):
        size = [size]

    if not isinstance(size, collections.Sequence):
        size = list(size)

    if not cov.shape == (d, d):
        raise ValueError('mean and cov must be a arrays with shapes (d, ) and '
                         '(d, d) respectively')

    draw_shape = list(size)
    draw_shape.append(d)
    n = int(np.prod(size))
    if mean.dtype != cov.dtype:
        raise ValueError('the dtypes of mean and cov mjust match')

    z = randn(n, d, dtype=cov.dtype)
    cov_cholesky = cholesky(cov).T
    x = z @ cov_cholesky

    return x.reshape(draw_shape)



import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 1500000
test_data = [(1, 2, n_kolmogorov_smirnov),
             (2, 2, n_kolmogorov_smirnov),
             (3, 2, n_kolmogorov_smirnov),
             (5, 1, n_kolmogorov_smirnov),
             (9, 0.5, n_kolmogorov_smirnov),
             (7.5, 1, n_kolmogorov_smirnov),
             (0.5, 1, n_kolmogorov_smirnov)]


@pytest.mark.parametrize("a, b, n_kolmogorov_smirnov", test_data)
def test_gamma_distribution(a, b, n_kolmogorov_smirnov):
    u = cn.random.gamma(a, b, n_kolmogorov_smirnov)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='gamma',
                             args=(a, 0.0, b),
                             verbose=True)

    assert not reject



import array as pyarray
import numpy as np
import pytest

import cocos.device
import cocos.numerics as cn


# backend = 'cpu'
# backend = 'cuda'
# backend = 'opencl'
backend = None


def test_concatenate_hstack_v_stack_dstack():
    cocos.device.init(backend)
    cocos.device.info()

    # define data type
    dtype = np.int32

    # using numpy
    first_array_numpy = np.array([[1, 2], [3, 4]], dtype=dtype)
    second_array_numpy = np.array([[5, 6], [7, 8]], dtype=dtype)
    third_array_numpy = np.array([[9, 10], [11, 12]], dtype=dtype)
    h_stacked_array_numpy = np.hstack((first_array_numpy,
                                       second_array_numpy,
                                       third_array_numpy))

    v_stacked_array_numpy = np.vstack((first_array_numpy,
                                       second_array_numpy,
                                       third_array_numpy))

    d_stacked_array_numpy = np.dstack((first_array_numpy,
                                       second_array_numpy,
                                       third_array_numpy))
    # print(d_stacked_array_numpy)

    # using Cocos
    first_array_cocos = cn.array(first_array_numpy)
    second_array_cocos = cn.array(second_array_numpy)
    third_array_cocos = cn.array(third_array_numpy)
    h_stacked_array_cocos = cn.hstack((first_array_cocos,
                                       second_array_cocos,
                                       third_array_cocos))

    v_stacked_array_cocos = cn.vstack((first_array_cocos,
                                       second_array_cocos,
                                       third_array_cocos))

    d_stacked_array_cocos = cn.dstack((first_array_cocos,
                                       second_array_cocos,
                                       third_array_cocos))
    # print(d_stacked_array_cocos)

    assert np.allclose(h_stacked_array_cocos, h_stacked_array_numpy)
    assert np.allclose(v_stacked_array_cocos, v_stacked_array_numpy)
    assert np.allclose(d_stacked_array_cocos, d_stacked_array_numpy)

    # tests concatenate
    for axis in range(2):
        concatenated_array = np.concatenate((first_array_numpy,
                                             second_array_numpy,
                                             third_array_numpy),
                                            axis=axis)

        concatenated_array_cocos = cn.concatenate((first_array_cocos,
                                                   second_array_cocos,
                                                   third_array_cocos),
                                                  axis=axis)

        assert np.allclose(concatenated_array_cocos, concatenated_array)



import collections
import arrayfire as af
import typing as tp


def convert_trans_to_af_matprop(trans: int) -> af.MATPROP:
    if trans == 0:
        return af.MATPROP.NONE
    elif trans == 1:
        return af.MATPROP.TRANS
    elif trans == 2:
        return af.MATPROP.CTRANS
    else:
        raise ValueError("trans must be 0, 1, or 2")


def is_broadcastable(shp1: tp.Tuple, shp2: tp.Tuple) -> bool:
    for a, b in zip(shp1[::-1], shp2[::-1]):
        if a == 1 or b == 1 or a == b:
            pass
        else:
            return False
    return True


def check_and_make_sequence(sequence_candidate,
                            item_class) -> tp.Sequence:
    if not isinstance(sequence_candidate, collections.Sequence):
        if isinstance(sequence_candidate, item_class):
            sequence_candidate = [sequence_candidate]
        else:
            raise TypeError("sequence_candidate must be of of type "
                            "Sequence[item_class]")

    return sequence_candidate



from enum import Enum


class MixedComputationErrorLevel(Enum):
    NONE = 1
    WARNING = 2
    ERROR = 3


class RandomNumberGenerator(Enum):
    PHILOX_4X32_10 = 1
    THREEFRY_2X32_16 = 2
    MERSENNE_GP11213 = 3
    PHILOX = 1
    THREEFRY = 2
    DEFAULT = 1


class GPUOptions:
    def __init__(self):
        print("running GPUOptions.__init__")
    default_rng = RandomNumberGenerator.DEFAULT
    mixed_computation_error_level = MixedComputationErrorLevel.ERROR
    str_via_numpy = True
    use_gpu = True
    verbose = False



import cocos.device
import cocos.numerics as cn
import cocos.scientific.linalg as csl
import numpy as np
import pytest
import scipy.linalg as scipy_linalg

test_data = [(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]],
                       dtype=np.float64),
             np.array([[100, 1000], [200, 2000], [300, 3000]],
                      dtype=np.float64))]


@pytest.mark.parametrize("A_numpy, b_numpy", test_data)
def test_solve(A_numpy, b_numpy):
    cocos.device.init()

    if not cocos.device.is_dbl_supported():
        A_numpy = A_numpy.astype(np.float32)
        b_numpy = b_numpy.astype(np.float32)

    # using numpy
    lu, piv = scipy_linalg.lu_factor(A_numpy,
                                     overwrite_a=False)

    solution = scipy_linalg.lu_solve((lu, piv),
                                     b_numpy,
                                     trans=0,
                                     overwrite_b=False)

    # using Cocos
    b_cocos = cn.array(b_numpy)

    piv_cocos, lu_cocos = csl._lu_internal(cn.array(A_numpy),
                                           permute_l=True,
                                           overwrite_a=True)

    solution_cocos = csl.lu_solve((lu_cocos, piv_cocos),
                                  b_cocos,
                                  trans=0,
                                  overwrite_b=False)

    lu_cocos_2, piv_cocos_2 = csl.lu_factor(cn.array(A_numpy),
                                            overwrite_a=True)

    solution_cocos_2 = csl.lu_solve((lu_cocos_2, piv_cocos_2),
                                    b_cocos,
                                    trans=0,
                                    overwrite_b=False)

    # conduct tests
    assert np.allclose(solution, solution_cocos)
    assert np.allclose(solution, solution_cocos_2)



import cocos.device
import cocos.numerics as cn
import numpy as np


def test_diff():
    cocos.device.init()

    if cocos.device.is_dbl_supported():
        A_numpy = np.random.randn(1000, 1000)
        A_cocos = cn.array(A_numpy)

        # # using numpy
        # mean_numpy = np.mean(A_numpy)
        #
        # # using Cocos

        # mean_cocos = cn.mean(A_cocos)

        # conduct tests

        # tests sum
        for n in range(1, 10):
            assert np.allclose(np.diff(A_numpy, n=n),
                               cn.diff(A_cocos, n=n))

            assert np.allclose(np.diff(A_numpy, n=n, axis=0),
                               cn.diff(A_cocos, n=n, axis=0))

            assert np.allclose(np.diff(A_numpy, n=n, axis=1),
                               cn.diff(A_cocos, n=n, axis=1))



import numpy as np
import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 1000000
test_data = [(5.0, np.sqrt(3.0) / np.pi, n_kolmogorov_smirnov),
            (5.0, 2.0, n_kolmogorov_smirnov),
            (9.0, 3.0, n_kolmogorov_smirnov),
            (9.0, 4.0, n_kolmogorov_smirnov),
            (6.0, 2.0, n_kolmogorov_smirnov),
            (2.0, 1.0, n_kolmogorov_smirnov)]


@pytest.mark.parametrize("loc, scale, n_kolmogorov_smirnov", test_data)
def test_logistic_distribution(loc, scale, n_kolmogorov_smirnov):
    u = cn.random.logistic(loc, scale, n_kolmogorov_smirnov)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='logistic',
                             args=(loc, scale),
                             verbose=True)

    assert not reject



import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 1000000
test_data = [
             (1.0, 1.0, n_kolmogorov_smirnov),
             (1.0, 0.2, n_kolmogorov_smirnov),
             (1.0, 3.0, n_kolmogorov_smirnov),
             (3.0, 1.0, n_kolmogorov_smirnov),
             (3.0, 0.2, n_kolmogorov_smirnov)
            ]


@pytest.mark.parametrize("mean, scale, n_kolmogorov_smirnov", test_data)
def test_wald_distribution(mean, scale, n_kolmogorov_smirnov):
    u = cn.random.wald(mean, scale, n_kolmogorov_smirnov)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='invgauss',
                             args=(mean, scale, 0),
                             verbose=True)

    assert not reject



import numpy as np
import scipy.linalg as scipy_linalg

import cocos.device
import cocos.numerics as cn
import cocos.numerics.linalg


def compare_cocos_numpy(cocos_array, numpy_array):
    return np.allclose(np.array(cocos_array), numpy_array)


def test_cholesky():
    cocos.device.init()

    A_numpy = np.array([[1. + 0.j,  0. - 2.j], [0. + 2.j,  5. + 0.j]],
                       dtype=np.complex64)

    L_numpy = np.linalg.cholesky(A_numpy)
    # print("L_numpy")
    # print(L_numpy)
    # print("")

    A_cocos = cn.array(A_numpy)
    L_cocos = cn.linalg.cholesky(A_cocos)
    # print("L_numpy cocos")
    # print(np.array(L_cocos))
    # print("")

    assert compare_cocos_numpy(L_cocos, L_numpy)



import arrayfire as af
import typing as tp


class DeviceMemoryInfo:
    def __init__(self,
                 allocated_buffers: int,
                 allocated_bytes: int,
                 locked_buffers: int,
                 locked_bytes: int) -> None:
        self._allocated_buffers = allocated_buffers
        self._allocated_bytes = allocated_bytes
        self._locked_buffers = locked_buffers
        self._locked_bytes = locked_bytes

    @property
    def allocated_buffers(self) -> int:
        return self._allocated_buffers

    @property
    def allocated_bytes(self) -> int:
        return self._allocated_bytes

    @property
    def locked_buffers(self) -> int:
        return self._locked_buffers

    @property
    def locked_bytes(self) -> int:
        return self._locked_bytes

    def __str__(self):
        return f"allocated buffers = {self.allocated_buffers}\n" \
               f"allocated bytes = {self.allocated_bytes}\n" \
               f"locked buffers = {self.locked_buffers}\n" \
               f"locked bytes = {self.locked_bytes}"


class ComputeDevice:
    def __init__(self,
                 id: int,
                 name: str,
                 backend: str,
                 toolkit_version: str,
                 compute_version: str) -> None:
        self._id = id
        self._name = name
        self._backend = backend
        self._toolkit_version = toolkit_version
        self._compute_version = compute_version

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def backend(self) -> str:
        return self._backend

    @property
    def toolkit_version(self) -> str:
        return self._toolkit_version

    @property
    def compute_version(self) -> str:
        return self._compute_version

    def sync(self):
        sync(self.id)

    def __str__(self):
        return f"{self.id}: {self.name} | " \
               f"{self.backend} - {self.toolkit_version} | " \
               f"compute version {self.compute_version}"


def get_current_device_memory_info() \
        -> DeviceMemoryInfo:
    device_mem_info = af.device_mem_info()
    allocated_buffers = device_mem_info['alloc']['buffers']
    allocated_bytes = device_mem_info['alloc']['bytes']
    locked_buffers = device_mem_info['lock']['buffers']
    locked_bytes = device_mem_info['lock']['bytes']

    return DeviceMemoryInfo(allocated_buffers=allocated_buffers,
                            allocated_bytes=allocated_bytes,
                            locked_buffers=locked_buffers,
                            locked_bytes=locked_bytes)


class ComputeDevicePool:
    def __init__(self,
                 compute_devices: tp.Sequence[tp.Union[int, ComputeDevice]]) \
            -> None:
        self._compute_devices = []

        for i, compute_device in enumerate(compute_devices):
            if isinstance(compute_device, int):
                compute_device \
                    = ComputeDeviceManager.get_compute_device(compute_device)
            elif not isinstance(compute_device, ComputeDevice):
                raise TypeError(f"Every element in compute_devices must be of "
                                f"type ComputeDevice or of type int. Entry {i} "
                                f"is of type {type(compute_device)}")

            self._compute_devices.append(compute_device)

        self._compute_devices: tp.Tuple[ComputeDevice] \
            = tuple(self._compute_devices)

    def compute_devices(self) -> tp.Tuple[ComputeDevice]:
        return self._compute_devices

    def sync(self):
        for compute_device in self._compute_devices:
            compute_device.sync()


class ComputeDeviceManager:
    _compute_devices = None

    @staticmethod
    def _get_compute_device_internal(id: int) -> ComputeDevice:
        af.set_device(id)
        device_info = af.device_info()
        name = device_info['device']
        backend = device_info['backend']
        toolkit_version = device_info['toolkit']
        compute_version = device_info['compute']

        return ComputeDevice(id,
                             name,
                             backend,
                             toolkit_version,
                             compute_version)

    @staticmethod
    def get_number_of_compute_devices() -> int:
        return len(ComputeDeviceManager.get_compute_devices())

    @classmethod
    def get_compute_devices(cls) -> tp.Sequence[ComputeDevice]:
        if ComputeDeviceManager._compute_devices is None:
            saved_device_id = cls.get_current_compute_device_id()
            n = af.get_device_count()
            ComputeDeviceManager._compute_devices = []

            for id in range(n):
                (ComputeDeviceManager
                 ._compute_devices
                 .append(ComputeDeviceManager._get_compute_device_internal(id)))

            af.set_device(saved_device_id)

        return ComputeDeviceManager._compute_devices

    @staticmethod
    def get_compute_devices_by_name(name_contains: str) \
            -> tp.Sequence[ComputeDevice]:
        compute_devices = ComputeDeviceManager.get_compute_devices()
        return [compute_device for compute_device in compute_devices
                if name_contains in compute_device.name]

    @staticmethod
    def get_compute_device(id: int) -> ComputeDevice:
        return ComputeDeviceManager.get_compute_devices()[id]

    @classmethod
    def get_current_compute_device_id(cls) -> int:
        return af.get_device()

    @classmethod
    def get_current_compute_device(cls) -> ComputeDevice:
        return cls.get_compute_device(cls.get_current_compute_device_id())

    @staticmethod
    def set_compute_device(compute_device: tp.Union[int, ComputeDevice]) \
            -> None:
        if isinstance(compute_device, int):
            compute_device \
                = ComputeDeviceManager.get_compute_device(compute_device)
        elif not isinstance(compute_device, ComputeDevice):
            raise TypeError(f"The argument compute_device must be of "
                            f"type ComputeDevice or of type int. The argument "
                            f"provided is of type {type(compute_device)}")

        af.set_device(compute_device.id)


def init(backend: tp.Optional[str] = None):
    try:
        if backend:
            af.set_backend(backend)
        af.get_device()
    except:
        af.set_backend('cpu')


def info():
    print("Cocos running on " + build_and_backend())
    selected_device_id = ComputeDeviceManager.get_current_compute_device_id()
    for compute_device in ComputeDeviceManager.get_compute_devices():
        device_id = compute_device.id
        if device_id == selected_device_id:
            device_string = "[" + str(device_id) + "]"
        else:
            device_string = "-" + str(device_id) + "-"
        device_name = compute_device.name.replace('_', ' ')
        print(
            f'{device_string} {compute_device.toolkit_version}: {device_name} '
            f'| {compute_device.backend} '
            f'| compute version {compute_device.compute_version}')


def build_and_backend() -> str:
    return af.info_str().split('\n')[0]


def sync(compute_device: tp.Optional[tp.Union[int, ComputeDevice]] = None):
    if isinstance(compute_device, ComputeDevice):
        device = compute_device.id
    elif not compute_device or isinstance(compute_device, int):
        device = compute_device
    else:
        raise TypeError(f"The argument compute_device must be of "
                        f"type ComputeDevice or of type int. The argument "
                        f"provided is of type {type(compute_device)}")

    af.sync(device=device)


def gpu_sync_wrapper(f: tp.Callable, *args, **kwargs):
    result = f(*args, **kwargs)
    sync()
    return result


def is_dbl_supported(
        compute_device: tp.Optional[tp.Union[int, ComputeDevice]] = None) \
        -> bool:
    if isinstance(compute_device, ComputeDevice):
        device = compute_device.id
    elif not compute_device or isinstance(compute_device, int):
        device = compute_device
    else:
        raise TypeError(f"The argument compute_device must be of "
                        f"type ComputeDevice or of type int. The argument "
                        f"provided is of type {type(compute_device)}")

    return af.is_dbl_supported(device)


def selected_backend() -> str:
    device_info = af.device_info()
    return device_info['backend'].lower()



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1 + 2j, 2, 3],
                       [4, 5 + 1j, 6],
                       [7, 8, 20 + 7j]],
                      dtype=np.complex64),
             np.array([[0, 2, 3],
                       [4, 0, 6],
                       [7, 8, 9]],
                      dtype=np.int32)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_nonzero(A_numpy):
    cocos.device.init()

    A_cocos = cn.array(A_numpy)

    # flatnonzero
    assert np.allclose(cn.flatnonzero(A_cocos), np.flatnonzero(A_numpy))

    # nonzero
    (nonzero_i, nonzero_j) = np.nonzero(A_numpy)
    (nonzero_i_self, nonzero_j_self) = A_numpy.nonzero()
    (nonzero_i_cocos, nonzero_j_cocos) = cn.nonzero(A_cocos)
    (nonzero_i_cocos_self, nonzero_j_cocos_self) = A_cocos.nonzero()
    # print("nonzero numpy i")
    # print(nonzero_i)
    # print("nonzero numpy j")
    # print(nonzero_j)
    # print("nonzero cocos i")
    # print(nonzero_i_cocos)
    # print("nonzero cocos i")
    # print(nonzero_j_cocos)
    assert np.allclose(nonzero_i_cocos, nonzero_i)
    assert np.allclose(nonzero_j_cocos, nonzero_j)
    assert np.allclose(nonzero_i_cocos_self, nonzero_i_self)
    assert np.allclose(nonzero_j_cocos_self, nonzero_j_self)




import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest

test_data = [np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             np.array([[0.5, 2.3, 3.1],
                       [4, 5.5, 6],
                       [7 - 9j, 8 + 1j, 2 + 10j]],
                      dtype=np.complex64)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_diag_trace(A_numpy):
    if A_numpy.ndim != 2:
        raise ValueError("A_numpy must be a matrix")
    if A_numpy.shape[0] != A_numpy.shape[1]:
        raise ValueError("A_numpy must be a square matrix")

    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    # conduct tests
    for k in range(-1, 2):
        assert np.allclose(np.diag(A_numpy, k=k), cn.diag(A_cocos, k=k))

    # tests diag and trace
    diag_numpy = np.diag(A_numpy)
    diag_cocos = cn.diag(A_cocos)

    assert np.allclose(np.diag(diag_numpy), cn.diag(diag_cocos))
    assert np.isclose(np.trace(A_numpy), cn.trace(A_cocos))

    # tests diagonal
    diagonal_numpy = A_numpy.diagonal()
    diagonal_cocos = A_cocos.diagonal()

    assert np.allclose(diagonal_numpy, diagonal_cocos)

    d = A_numpy.shape[0]

    for offset in range(-d+1, d):
        diagonal_numpy = A_numpy.diagonal(offset=offset)
        diagonal_cocos = A_cocos.diagonal(offset=offset)
        assert np.allclose(diagonal_numpy, diagonal_cocos)

    # tests diagflat
    assert np.allclose(np.diagflat(diag_numpy), cn.diagflat(diag_cocos))



import arrayfire as af
import numpy as np
import typing as tp

from cocos.numerics._array import ndarray, isvector
from cocos.numerics._data import zeros
from cocos.utilities import convert_trans_to_af_matprop


def cholesky(a: ndarray,
             overwrite: bool = False,
             is_upper: bool = False) -> ndarray:
    """
    Compute the Cholesky decomposition of a matrix.
    """

    if overwrite:
        info = af.cholesky_inplace(a._af_array, is_upper=is_upper)
        if info == 0:
            return a
        else:
            raise np.linalg.LinAlgError(f"cholesky decomposition failed at "
                                        f"rank {info}")
    else:
        R, info = af.cholesky(a._af_array, is_upper=is_upper)
        if info == 0:
            return ndarray(R)
        else:
            raise np.linalg.LinAlgError(f"cholesky decomposition failed at "
                                        f"rank {info}")


def qr(a: ndarray,
       mode: str = 'reduced',
       overwrite: bool = False) \
        -> tp.Union[tp.Tuple[ndarray, ndarray, ndarray], ndarray]:
    """
    Compute QR decomposition of a matrix.
    """

    if overwrite:
        return ndarray(af.qr_inplace(a._af_array))
    else:
        (Q, R, T) = af.qr(a._af_array)
        return ndarray(Q), ndarray(R), ndarray(T)


def svd(a: ndarray,
        full_matrices: bool = True,
        compute_uv: bool = True,
        overwrite: bool = False) \
        -> tp.Union[tp.Tuple[ndarray, ndarray, ndarray], ndarray]:
    """
    Singular Value Decomposition.
    """

    if not full_matrices:
        raise ValueError("full_matrices must be True")

    if overwrite:
        U, S, Vt = af.svd_inplace(a._af_array)
    else:
        U, S, Vt = af.svd(a._af_array)

    if compute_uv:
        return ndarray(U), ndarray(S), ndarray(Vt)
    else:
        return ndarray(S)


def _convert_pivot_matrix(pivot: ndarray):
    if not isvector(pivot):
        raise ValueError("pivot must be a vector")

    n = pivot.size
    out = zeros((n, n), dtype=np.int32)

    pivot_np = np.array(pivot)

    for i in range(n):
        out[pivot_np[i], i] = 1

    return out


def lu(a: np.ndarray,
       permute_l: bool = False,
       overwrite_a: bool = False,
       check_finite: bool = True,
       is_lapack_piv: bool = True):
    """
    Compute pivoted LU decomposition of a matrix.
    """

    if overwrite_a:
        pivot, LU = _lu_internal(a,
                                 permute_l=permute_l,
                                 overwrite_a=overwrite_a,
                                 is_lapack_piv=is_lapack_piv)

        return _convert_pivot_matrix(pivot), LU
    else:
        pivot, lower, upper = _lu_internal(a,
                                           permute_l=permute_l,
                                           overwrite_a=overwrite_a)

        return _convert_pivot_matrix(pivot), lower, upper


def lu_factor(a: ndarray,
              overwrite_a: bool = False,
              check_finite: bool = True) -> tp.Tuple[ndarray, ndarray]:
    """
    Compute pivoted LU decomposition of a matrix.
    """

    if not overwrite_a:
        raise ValueError("overwrite_a != True is not supported")

    pivot, LU = _lu_internal(a,
                             permute_l=True,
                             overwrite_a=overwrite_a,
                             is_lapack_piv=True)

    return LU, pivot


def _lu_internal(a: ndarray,
                 permute_l: bool = False,
                 overwrite_a: bool = False,
                 is_lapack_piv: bool = True) \
        -> tp.Union[tp.Tuple[ndarray, ndarray],
                    tp.Tuple[ndarray, ndarray, ndarray]]:
    if overwrite_a:
        if permute_l:
            if is_lapack_piv:
                pivot = "lapack"
            else:
                pivot = "full"

            P = af.lu_inplace(a._af_array, pivot=pivot)
            return ndarray(P), a
        else:
            raise ValueError("cannot overwrite a without permuting l")
    else:
        if permute_l:
            raise ValueError("cannot permute l without overwriting a")
        else:
            (L, U, P) = af.lu(a._af_array)
            return ndarray(P), ndarray(L), ndarray(U)


def lu_solve(lu_and_piv: tp.Tuple[ndarray, ndarray],
             b: ndarray,
             trans: int = 0,
             overwrite_b: bool = False,
             check_finite: bool = True) -> ndarray:
    """
    Solve an equation system, a x = b, given the LU factorization of a
    """

    if len(lu_and_piv) != 2:
        raise ValueError("lu_and_piv must be a two-dimensional tuple")

    lu, piv = lu_and_piv
    options = convert_trans_to_af_matprop(trans)

    return ndarray(af.solve_lu(lu._af_array,
                               piv._af_array,
                               b._af_array,
                               options=options))



import cocos.device
import cocos.numerics as cn
import cocos.numerics.linalg as linalg
import numpy as np


def compare_cocos_numpy(cocos_array, numpy_array):
    return np.allclose(np.array(cocos_array), numpy_array)


def test_qr():
    cocos.device.init()

    A_numpy = np.array([[1 + 2j, 2, 3], [4, 5 + 1j, 6], [7, 8, 20 + 7j]],
                       dtype=np.complex64)

    Q, R = np.linalg.qr(A_numpy)
    h, tau = np.linalg.qr(A_numpy, 'raw')
    print("Q numpy")
    print(Q)
    print("")

    print("R numpy")
    print(R)
    print("")

    print("tau numpy")
    print(tau)
    print("")

    A_cocos = cn.array(A_numpy)
    print(A_cocos.dtype)
    Q_cocos, R_cocos, T_cocos = cn.linalg.qr(A_cocos)
    print("Q cocos")
    print(np.array(Q_cocos))
    print("")

    print("R cocos")
    print(np.array(R_cocos))
    print("")

    print("tau cocos")
    print(np.array(T_cocos))
    print("")

    assert compare_cocos_numpy(Q_cocos, Q)
    assert compare_cocos_numpy(R_cocos, R)
    assert compare_cocos_numpy(T_cocos, tau)



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.3, 0.25]],
                      dtype=np.float32)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_argmin(A_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    # tests argmin
    assert np.allclose(np.argmin(A_numpy.transpose()), cn.argmin(A_cocos))
    assert np.allclose(np.argmin(A_numpy, axis=0), cn.argmin(A_cocos, axis=0))
    assert np.allclose(np.argmin(A_numpy, axis=1), cn.argmin(A_cocos, axis=1))

    # tests argmax
    print("numpy argmax")
    print(np.argmax(A_numpy.transpose()))
    print("cocos argmax")
    print(cn.argmax(A_cocos))
    assert np.allclose(np.argmax(A_numpy.transpose()), cn.argmax(A_cocos))
    assert np.allclose(np.argmax(A_numpy, axis=0), cn.argmax(A_cocos, axis=0))
    assert np.allclose(np.argmax(A_numpy, axis=1), cn.argmax(A_cocos, axis=1))



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest
import scipy.special as sp


test_data = [np.array([[np.inf, -np.inf, np.NaN]],
                      dtype=np.float32),
             np.array([[0, 0]],
                      dtype=np.int32),
             np.array([[1, 2, -3],
                       [4, -5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5],
                       [0.4, 0.5, 0.6],
                       [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             np.array([[0.5, 2.3, 3.1],
                       [4, 5.5, 6],
                       [7 - 9j, 8 + 1j, 2 + 10j]],
                      dtype=np.complex64)]


@pytest.mark.parametrize("A_numpy", test_data)
def test_unary(A_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)

    all_positive = np.all(A_numpy > 0)

    all_finite = cn.all(cn.isfinite(A_cocos))
    assert np.allclose(np.isfinite(A_numpy), cn.isfinite(A_cocos))

    if all_finite:
        if cn.isrealobj(A_cocos):
            if np.count_nonzero(A_numpy) == \
                    A_numpy.size and cn.isfloating(A_cocos):
                assert np.allclose(1 / A_numpy, 1 / A_cocos)
            if all_positive:
                assert np.allclose(np.sqrt(A_numpy), cn.sqrt(A_cocos))
                assert np.allclose(np.cbrt(A_numpy), cn.cbrt(A_cocos))

                # af.root does not seem to work
                # if cn.isfloating(A_cocos):
                #     print("cbrt numpy")
                #     print(np.cbrt(A_numpy))
                #     print("third root cocos")
                #     print(cn.root(A_cocos, 3))
                #     print("cbrt cocos")
                #     print(cn.cbrt(A_cocos))
                #
                #     assert np.allclose(np.cbrt(A_numpy), cn.root(A_cocos, 3))
                assert np.allclose(np.log(A_numpy), cn.log(A_cocos))
                assert np.allclose(np.log1p(A_numpy), cn.log1p(A_cocos))
                assert np.allclose(np.log2(A_numpy), cn.log2(A_cocos))
                assert np.allclose(np.log10(A_numpy), cn.log10(A_cocos))
                assert np.allclose(sp.gamma(A_numpy), cn.gamma(A_cocos))
                assert np.allclose(sp.gammaln(A_numpy), cn.gammaln(A_cocos))
                assert np.allclose(np.power(2, A_numpy), cn.power2(A_cocos))

                if cn.isinteger(A_cocos):
                    assert np.allclose(sp.factorial(A_numpy),
                                       cn.factorial(A_cocos))

            # assert np.allclose(np.round(A), cn.round(A_cocos))
            assert np.allclose(np.isneginf(A_numpy), cn.isneginf(A_cocos))
            assert np.allclose(np.isposinf(A_numpy), cn.isposinf(A_cocos))
            assert np.allclose(np.trunc(A_numpy), cn.trunc(A_cocos))
            assert np.allclose(np.floor(A_numpy), cn.floor(A_cocos))
            assert np.allclose(np.ceil(A_numpy), cn.ceil(A_cocos))
            assert np.allclose(np.sign(A_numpy), cn.sign(A_cocos))
            assert np.allclose(np.sinh(A_numpy), cn.sinh(A_cocos))
            assert np.allclose(np.cosh(A_numpy), cn.cosh(A_cocos))
            assert np.allclose(np.tanh(A_numpy), cn.tanh(A_cocos))
            assert np.allclose(np.sin(A_numpy), cn.sin(A_cocos))
            assert np.allclose(np.cos(A_numpy), cn.cos(A_cocos))
            assert np.allclose(np.tan(A_numpy), cn.tan(A_cocos))
            assert np.allclose(sp.expit(A_numpy), cn.expit(A_cocos))
            assert np.allclose(np.expm1(A_numpy), cn.expm1(A_cocos))
            assert np.allclose(sp.erf(A_numpy), cn.erf(A_cocos))
            assert np.allclose(sp.erfc(A_numpy), cn.erfc(A_cocos))

            if cn.isinteger(A_cocos):
                # print("invert numpy")
                # print(np.invert(A))
                # print("invert cocos")
                # print(cn.invert(A_cocos))
                assert np.allclose(np.invert(A_numpy), cn.invert(A_cocos))
        else:
            assert np.allclose(np.real(A_numpy), cn.real(A_cocos))
            assert np.allclose(np.imag(A_numpy), cn.imag(A_cocos))
            assert np.allclose(np.conj(A_numpy), cn.conj(A_cocos))
            assert np.allclose(np.angle(A_numpy), cn.angle(A_cocos))

        if not all_positive:
            np.allclose(np.absolute(A_numpy), cn.absolute(A_cocos))

        assert np.allclose(np.exp(A_numpy), cn.exp(A_cocos))
        assert np.allclose(A_numpy == 0, cn.iszero(A_cocos))

    assert np.allclose(np.isfinite(A_numpy), cn.isfinite(A_cocos))
    assert np.allclose(np.isinf(A_numpy), cn.isinf(A_cocos))
    assert np.allclose(np.isnan(A_numpy), cn.isnan(A_cocos))



import pytest

import cocos.numerics as cn
from cocos.tests.test_numerics.test_statistics.utilities import perform_ks_test


n_kolmogorov_smirnov = 5000000
test_data = [( 0.0, 0.2, n_kolmogorov_smirnov),
             ( 0.0, 1.0, n_kolmogorov_smirnov),
             ( 0.0, 5.0, n_kolmogorov_smirnov),
             (-2.0, 0.5, n_kolmogorov_smirnov)]


@pytest.mark.parametrize("mu, sigma, n_kolmogorov_smirnov", test_data)
def test_normal_distribution(mu, sigma, n_kolmogorov_smirnov):
    u = cn.random.normal(mu, sigma, n_kolmogorov_smirnov)
    reject = perform_ks_test(u,
                             alpha=0.01,
                             distribution='norm',
                             args=(mu, sigma),
                             verbose=True)

    assert not reject



import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data_int = [(np.array([[1, 2, -3],
                            [4, -5, 6],
                            [7, 8, 20]],
                           dtype=np.int32),
                  np.array([[2, 1, 5],
                            [4, 5, 6],
                            [7, 2, 25]],
                           dtype=np.int32))]


test_data_float = [(np.array([[1, 2, -3],
                              [4, -5, 6],
                              [7, 8, 20]],
                             dtype=np.float32),
                    np.array([[2, 1, 5],
                              [4, 5, 6],
                              [7, 2, 25]],
                             dtype=np.float32))]


@pytest.mark.parametrize("A_numpy, B_numpy", test_data_float)
def test_binary_float(A_numpy, B_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)
    B_cocos = cn.array(B_numpy)

    assert np.allclose(A_cocos + B_cocos, A_numpy + B_numpy)
    assert np.allclose(A_cocos - B_cocos, A_numpy - B_numpy)
    assert np.allclose(A_cocos * B_cocos, A_numpy * B_numpy)
    assert np.allclose(A_cocos / B_cocos, A_numpy / B_numpy)
    assert np.allclose(cn.hypot(A_cocos, B_cocos),
                       np.hypot(A_numpy, B_numpy))

    assert np.allclose(cn.arctan2(A_cocos, B_cocos),
                       np.arctan2(A_numpy, B_numpy))

    assert np.allclose(cn.cplx(A_cocos, B_cocos),
                       A_numpy + 1j * B_numpy)

    assert np.allclose(np.power(A_numpy, B_numpy),
                       cn.power(A_cocos, B_cocos))


@pytest.mark.parametrize("A_numpy, B_numpy", test_data_int)
def test_binary_int(A_numpy, B_numpy):
    cocos.device.init()
    A_cocos = cn.array(A_numpy)
    B_coocs = cn.array(B_numpy)

    assert np.allclose(cn.bitwise_and(A_cocos, B_coocs),
                       np.bitwise_and(A_numpy, B_numpy))

    assert np.allclose(cn.bitwise_or(A_cocos, B_coocs),
                       np.bitwise_or(A_numpy, B_numpy))

    assert np.allclose(cn.bitwise_xor(A_cocos, B_coocs),
                       np.bitwise_xor(A_numpy, B_numpy))

    assert np.allclose(cn.minimum(A_cocos, B_coocs),
                       np.minimum(A_numpy, B_numpy))

    assert np.allclose(cn.maximum(A_cocos, B_coocs),
                       np.maximum(A_numpy, B_numpy))

    assert np.allclose(cn.left_shift(A_cocos, B_coocs),
                       np.left_shift(A_numpy, B_numpy))

    assert np.allclose(cn.right_shift(A_cocos, B_coocs),
                       np.right_shift(A_numpy, B_numpy))

    assert np.allclose(A_cocos + B_coocs, A_numpy + B_numpy)
    assert np.allclose(A_cocos - B_coocs, A_numpy - B_numpy)
    assert np.allclose(A_cocos * B_coocs, A_numpy * B_numpy)



import arrayfire as af

from numbers import Number
import numpy as np
import typing as tp

from cocos.numerics._array import ndarray, isvector
from cocos.numerics._array import _unary_function
from cocos.numerics._data import zeros
from cocos.scientific.linalg import cholesky as chol_sc
from cocos.scientific.linalg import qr as qr_sc
from cocos.scientific.linalg import svd as svd_sc
from cocos.utilities import convert_trans_to_af_matprop


def cholesky(a: ndarray) -> ndarray:
    """
    Cholesky decomposition.
    """

    return chol_sc(a)


def qr(a: ndarray) -> tp.Union[tp.Tuple[ndarray, ndarray, ndarray], ndarray]:
    """
    Compute the qr factorization of a matrix."""

    return qr_sc(a)


def svd(a: ndarray,
        full_matrices: bool=True,
        compute_uv: bool=True) \
        -> tp.Union[tp.Tuple[ndarray, ndarray, ndarray], ndarray]:
    """
    Singular Value Decomposition.
    """

    return svd_sc(a, full_matrices, compute_uv)


def solve(a: ndarray,
          b: ndarray,
          trans: int = 0) -> ndarray:
    """
    Solve a linear matrix equation, or system of linear scalar equations.
    """

    options = convert_trans_to_af_matprop(trans)
    return ndarray(af.solve(a._af_array, b._af_array, options=options))


def inv(a: ndarray) -> ndarray:
    """
    Compute the (multiplicative) inverse of a matrix.
    """

    return _unary_function(a, af.inverse)


def matrix_rank(M: ndarray,
                tol: tp.Optional[float]=None) -> int:
    """
    Return matrix rank of array using SVD method
    """

    if tol is None:
        tol = 1e-5
    elif not isinstance(tol, Number):
        raise TypeError("tol must be numeric")

    return af.rank(M._af_array, tol)


def det(a: ndarray) -> float:
    """
    Compute the determinant of an array.
    """

    return af.det(a._af_array)


def norm(x: ndarray,
         ord: tp.Optional[int]=None,
         axis: tp.Optional[int]=None,
         keepdims: bool = False) -> float:
    """
    Matrix or vector norm.
    """

    if axis is not None:
        raise ValueError("axis != None is not supported")

    p = 1.0
    q = 1.0

    if ord == 2:
        S = svd(x, compute_uv=False)
        return S.max()
    else:
        if ord is None:
            norm_type = af.NORM.VECTOR_2
        elif isinstance(ord, Number):
            if ord == float("inf") or ord == np.inf:
                norm_type = af.NORM.MATRIX_INF
            elif ord == 1:
                norm_type = af.NORM.MATRIX_1
            else:
                norm_type = af.NORM.VECTOR_P
                p = ord
        else:
            raise ValueError(f"ord = {ord} is not supported")

        return af.norm(x._af_array, norm_type=norm_type, p=p, q=q)



from cocos.numerics.random import get_antithetic_slices


def test_antithetic_slices():
    expected_output_1 = (slice(0, 1, 1), slice(0, 10, 1))
    output_1 = get_antithetic_slices((2, 10), antithetic_dimension=0)
    assert output_1 == expected_output_1

    expected_output_2 = (slice(0, 2, 1), slice(0, 5, 1))
    output_2 = get_antithetic_slices((2, 10), antithetic_dimension=1)
    assert output_2 == expected_output_2

    expected_output_3 = (slice(0, 2, 1), slice(0, 4, 1))
    output_3 = get_antithetic_slices((2, 9), antithetic_dimension=1)
    assert output_3 == expected_output_3


import cocos.device
import cocos.numerics as cn
import numpy as np
import pytest


test_data = [np.array([[1, 2, 3], [4, 5, 6], [7, 8, 20]],
                      dtype=np.int32),
             np.array([[0.2, 1.0, 0.5], [0.4, 0.5, 0.6], [0.7, 0.2, 0.25]],
                      dtype=np.float32),
             np.array([[0.5, 2.3, 3.1], [4, 5.5, 6], [7 - 9j, 8 + 1j, 2 + 10j]],
                      dtype=np.complex64)]


@pytest.mark.parametrize("A", test_data)
def test_mean(A):
    cocos.device.init()
    A_arch = cn.array(A)

    # # using numpy
    # mean_numpy = np.mean(A)
    #
    # # using Archimedes

    # mean_arch = cn.mean(A_arch)

    # conduct tests

    # tests mean
    assert np.allclose(np.mean(A), cn.mean(A_arch))
    assert np.allclose(np.mean(A, axis=0), cn.mean(A_arch, axis=0))
    assert np.allclose(np.mean(A, axis=1), cn.mean(A_arch, axis=1))



import arrayfire as af
from arrayfire.arith import _arith_unary_func, _arith_binary_func
from arrayfire.array import _binary_func
from arrayfire.library import backend
import numbers
import numpy as np
import typing as tp

from ._array import ndarray
from ._array import _binary_function, _unary_function
from ._array import ndarray, isinteger
from ._data import full, zeros


def absolute(x: ndarray) -> ndarray:
    """
    Calculate the absolute value element-wise.
    """

    return _unary_function(x, af.abs)


abs = absolute


@af.broadcast
def minimum(x1: tp.Union[float, ndarray],
            x2: tp.Union[float, ndarray]) \
        -> ndarray:
    """
    Element-wise minimum of array elements.
    """

    return _binary_function(x1, x2, af.minof)


@af.broadcast
def maximum(x1: tp.Union[float, ndarray],
            x2: tp.Union[float, ndarray]) \
        -> ndarray:
    """
    Element-wise maximum of array elements.
    """

    return _binary_function(x1, x2, af.maxof)


def sqrt(x: tp.Union[float, ndarray]) \
        -> ndarray:
    """
    Return the positive square-root of an array, element-wise.
    """

    return _unary_function(x, af.sqrt)


def square(x: ndarray) -> ndarray:
    """
    Return the element-wise square of the input.
    """

    return x * x


def exp(x: tp.Union[numbers.Number, ndarray]) -> ndarray:
    """
    Calculate the exponential of all elements in the input array.
    """

    return _unary_function(x, af.exp)


def log(x: tp.Union[float, ndarray]) -> ndarray:
    """
    Natural logarithm, element-wise.
    """

    return _unary_function(x, af.log)


def iszero(x: ndarray) -> ndarray:
    return _unary_function(x, af.iszero)


def isinf(x: ndarray) -> ndarray:
    return _unary_function(x, af.isinf)


def isnan(x: ndarray) -> ndarray:
    return _unary_function(x, af.isnan)


@af.broadcast
def logical_and(x1: ndarray,
                x2: ndarray):
    new_af_array \
        = _binary_func(x1._af_array,
                       x2._af_array,
                       backend.get().af_and)

    return ndarray(new_af_array)


@af.broadcast
def logical_or(x1: ndarray,
               x2: ndarray):
    new_af_array = _binary_func(x1._af_array,
                                x2._af_array,
                                backend.get().af_or)

    return ndarray(new_af_array)


def logical_not(x: ndarray):
    new_af_array = _arith_unary_func(x._af_array,
                                     backend.get().af_not)

    return ndarray(new_af_array)


@af.broadcast
def logical_xor(x1: ndarray,
                x2: ndarray):
    return bitwise_xor(x1.astype(np.bool8), x2.astype(np.bool8))
    # return logical_and(logical_or(a,b),(logical_not(logical_and(a, b))))


@af.broadcast
def bitwise_and(x1: ndarray,
                x2: ndarray):
    new_af_array \
        = _binary_func(x1._af_array, x2._af_array, backend.get().af_bitand)
    return ndarray(new_af_array)


@af.broadcast
def bitwise_or(x1: ndarray,
               x2: ndarray):
    new_af_array \
        = _binary_func(x1._af_array, x2._af_array, backend.get().af_bitor)
    return ndarray(new_af_array)


@af.broadcast
def invert(x: ndarray):
    if not isinteger(x):
        raise TypeError("invert only supportd integers arguments")
    af_array = x._af_array
    return ndarray(_binary_func(af_array, -1, backend.get().af_bitxor))


@af.broadcast
def bitwise_xor(x1: ndarray,
                x2: ndarray):
    new_af_array \
        = _binary_func(x1._af_array, x2._af_array, backend.get().af_bitxor)
    return ndarray(new_af_array)
    # return a.__xor__(b)


@af.broadcast
def left_shift(x1: ndarray,
               x2: ndarray):
    return x1 << x2


@af.broadcast
def right_shift(x1: ndarray,
                x2: ndarray):
    return x1 >> x2


@af.broadcast
def remainder(x1: ndarray,
              x2: ndarray):
    """
    Return element-wise remainder of division.
    """

    if isinstance(x1, ndarray):
        x1 = x1._af_array
    if isinstance(x2, ndarray):
        x2 = x2._af_array

    return ndarray(af.rem(x1, x2))


@af.broadcast
def mod(x1: ndarray,
        x2: ndarray):
    """
    Return element-wise remainder of division.
    """

    # return _binary_function(a, b, af.rem)
    new_af_array \
        = _arith_binary_func(x1._af_array, x2._af_array, backend.get().af_mod)

    return ndarray(new_af_array)


def angle(z: ndarray):
    """
    Return the angle of the complex argument.
    """

    return _unary_function(z, af.arg)


def sign(x: ndarray):
    """
    Returns an element-wise indication of the sign of a number.
    """

    intermediate = _unary_function(x, af.sign)
    z = zeros(x.shape, dtype=x.dtype)
    result = (x > z) - intermediate

    return result


def trunc(a: ndarray):
    """
    Return the truncated value of the input, element-wise.
    """

    return _unary_function(a, af.trunc)


def ceil(a: ndarray):
    """
    Return the ceiling of the input, element-wise.
    """

    return _unary_function(a, af.ceil)


@af.broadcast
def hypot(x1: ndarray,
          x2: ndarray):
    """
    Given the “legs” of a right triangle, return its hypotenuse.
    """

    return _binary_function(x1, x2, af.hypot)


def sin(x: ndarray):
    """
    Trigonometric sine, element-wise.
    """

    return _unary_function(x, af.sin)


def cos(x: ndarray):
    """
    Cosine element-wise.
    """

    return _unary_function(x, af.cos)


def tan(x: ndarray):
    """
    Compute tangent element-wise.
    """

    return _unary_function(x, af.tan)


def arcsin(x: ndarray):
    """
    Inverse sine, element-wise.
    """

    return _unary_function(x, af.asin)


def arccos(x: ndarray):
    """
    Trigonometric inverse cosine, element-wise.
    """

    return _unary_function(x, af.acos)


def arctan(x: ndarray):
    """
    Trigonometric inverse tangent, element-wise.
    """

    return _unary_function(x, af.atan)


@af.broadcast
def arctan2(a: ndarray,
            b: ndarray):
    """
    Element-wise arc tangent of x1/x2 choosing the quadrant correctly.
    """

    return _binary_function(a, b, af.atan2)


@af.broadcast
def cplx(a: ndarray, b: tp.Optional[ndarray] = None):
    return _binary_function(a, b, af.cplx)


def sinh(a: ndarray):
    """
    Hyperbolic sine, element-wise.
    """

    return _unary_function(a, af.sinh)


def cosh(a: ndarray):
    """
    Hyperbolic cosine, element-wise.
    """

    return _unary_function(a, af.cosh)


def tanh(a: ndarray):
    """
    Compute hyperbolic tangent element-wise.
    """

    return _unary_function(a, af.tanh)


def arcsinh(a: ndarray):
    """
    Inverse hyperbolic sine element-wise.
    """

    return _unary_function(a, af.asinh)


def arccosh(a: ndarray):
    """
    Inverse hyperbolic cosine, element-wise.
    """

    return _unary_function(a, af.acosh)


def arctanh(a: ndarray):
    """
    Inverse hyperbolic tangent element-wise.
    """

    return _unary_function(a, af.atanh)


@af.broadcast
def root(a: ndarray,
         b: ndarray):
    return _binary_function(a, b, af.root)


@af.broadcast
def power(x1: ndarray,
          x2: ndarray):
    """
    First array elements raised to powers from second array, element-wise.
    """

    return _binary_function(x1, x2, af.pow)


def power2(a: ndarray):
    return _unary_function(a, af.pow2)


def expit(a: ndarray):
    return _unary_function(a, af.sigmoid)


def expm1(x: ndarray):
    """
    Calculate exp(x) - 1 for all elements in the array.
    """

    return _unary_function(x, af.expm1)


def erf(a: ndarray):
    return _unary_function(a, af.erf)


def erfc(a: ndarray):
    return _unary_function(a, af.erfc)


def log1p(a: ndarray):
    """
    Return the natural logarithm of one plus the input array,
    element-wise.
    """

    return _unary_function(a, af.log1p)


def log10(a: ndarray):
    """
    Return the base 10 logarithm of the input array, element-wise.
    """

    return _unary_function(a, af.log10)


def log2(a: ndarray):
    """
    Base-2 logarithm of x.
    """

    return _unary_function(a, af.log2)


def cbrt(x: ndarray):
    """
    Return the cube-root of an array, element-wise.
    """

    return _unary_function(x, af.cbrt)


def factorial(a: ndarray):
    return _unary_function(a, af.factorial)


def gamma(a: ndarray):
    return _unary_function(a, af.tgamma)


def gammaln(a: ndarray):
    return _unary_function(a, af.lgamma)


def isfinite(a: ndarray):
    inf = isinf(a)
    nan = isnan(a)
    return logical_not(logical_or(inf, nan))


# @af.broadcast
def isneginf(a: ndarray):
    neginf = full((1,), np.NINF, a.dtype)
    return a == neginf


# @af.broadcast
def isposinf(a: ndarray):
    posinf = full((1,), np.inf, a.dtype)
    return a == posinf
